# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Neuron groups
'''
__all__ = ['NeuronGroup', 'linked_var']

from numpy import *
from scipy import rand, linalg, random
from numpy.random import exponential, randint
import copy
from units import *
from threshold import *
import bisect
from reset import *
from clock import *
from stateupdater import *
from inspection import *
from operator import isSequenceType
import types
from utils.circular import *
import magic
from itertools import count
from equations import *
from globalprefs import *
import sys
from brian_unit_prefs import bup
import numpy
from base import *
from group import *
from threshold import select_threshold
from collections import defaultdict

timedarray = None # ugly hack: import this module when it is needed, can't do it here because of order of imports
network = None # ugly hack: import this module when it is needed, can't do it here because of order of imports


class TArray(numpy.ndarray):
    '''
    This internal class is just used for when Brian sends an array t
    to an object. All the elements will be the same in this case, and
    you can check for isinstance(arr, TArray) to do optimisations based
    on this. This behaviour may change in the future.
    '''
    def __new__(subtype, arr):
        # All numpy.ndarray subclasses need something like this, see
        # http://www.scipy.org/Subclasses
        return numpy.array(arr, copy=False).view(subtype)


class LinkedVar(object):
    def __init__(self, source, var=0, func=None, when='start', clock=None):
        self.source = source
        self.var = var
        self.func = func
        self.when = when
        self.clock = clock

def linked_var(source, var=0, func=None, when='start', clock=None):
    """
    Used for linking one :class:`NeuronGroup` variable to another.
    
    Sample usage::
    
        G = NeuronGroup(...)
        H = NeuronGroup(...)
        G.V = linked_var(H, 'W')

    In this scenario, the variable V in group G will always be updated with
    the values from variable W in group H. The groups G and H must be the
    same size (although subgroups can be used if they are not the same size).
    
    Arguments:
    
    ``source``
        The group from which values will be taken.
    ``var``
        The state variable of the source group to take values from.
    ``func``
        An additional function of one argument to pass the source variable
        values through, e.g. ``func=lambda x:clip(x,0,Inf)`` to half rectify the
        values.
    ``when``
        The time in the main Brian loop at which the copy operation is performed,
        as explained in :class:`Network`.
    ``clock``
        The update clock for the copy operation, by default it will use the clock
        of the target group.
    """
    return LinkedVar(source, var, func, when, clock)


class NeuronGroup(magic.InstanceTracker, ObjectContainer, Group):
    """Group of neurons
    
    Initialised with arguments:
    
    ``N``
        The number of neurons in the group.
    ``model``
        An object defining the neuron model. It can be
        an :class:`Equations` object, a string defining an :class:`Equations` object,
        a :class:`StateUpdater` object, or a list or tuple of :class:`Equations` and
        strings.
    ``threshold=None``
        A :class:`Threshold` object, a function, a scalar quantity or a string.
        If ``threshold`` is a function with one argument, it will be
        converted to a :class:`SimpleFunThreshold`, otherwise it will be a
        :class:`FunThreshold`. If ``threshold`` is a scalar, then a constant
        single valued threshold with that value will be used. In this case,
        the variable to apply the threshold to will be guessed. If there is
        only one variable, or if you have a variable named one of
        ``V``, ``Vm``, ``v`` or ``vm`` it will be used. If ``threshold`` is a
        string then the appropriate threshold type will be chosen, for example
        you could do ``threshold='V>10*mV'``. The string must be a one line
        string.
    ``reset=None``
        A :class:`Reset` object, a function, a scalar quantity or a string. If it's a
        function, it will be converted to a :class:`FunReset` object. If it's
        a scalar, then a constant single valued reset with that value will
        be used. In this case,
        the variable to apply the reset to will be guessed. If there is
        only one variable, or if you have a variable named one of
        ``V``, ``Vm``, ``v`` or ``vm`` it will be used. If ``reset`` is a
        string it should be a series of expressions which are evaluated for
        each neuron that is resetting. The series of expressions can be
        multiline or separated by a semicolon. For example,
        ``reset=`Vt+=5*mV; V=Vt'``. Statements involving ``if`` constructions
        will often not work because the code is automatically vectorised.
        For such constructions, use a function instead of a string.
    ``refractory=0*ms``, ``min_refractory``, ``max_refractory``
        A refractory period, used in combination with the ``reset`` value
        if it is a scalar. For constant resets only, you can specify refractory
        as an array of length the number of elements in the group, or as a
        string, giving the name of a state variable in the group. In the case
        of these variable refractory periods, you should specify
        ``min_refractory`` (optional) and ``max_refractory`` (required).
    ``clock``
        A clock to use for scheduling this :class:`NeuronGroup`, if omitted the
        default clock will be used.
    ``order=1``
        The order to use for nonlinear differential equation solvers.
        TODO: more details.
    ``implicit=False``
        Whether to use an implicit method for solving the differential
        equations. TODO: more details.
    ``max_delay=0*ms``
        The maximum allowable delay (larger values use more memory).
        This doesn't usually need to be specified because Connections will update it.
    ``compile=False``
        Whether or not to attempt to compile the differential equation
        solvers (into Python code). Typically, for best performance, both ``compile``
        and ``freeze`` should be set to ``True`` for nonlinear differential equations.
    ``freeze=False``
        If True, parameters are replaced by their values at the time
        of initialization.
    ``method=None``
        If not None, the integration method is forced. Possible values are
        linear, nonlinear, Euler, exponential_Euler (overrides implicit and order
        keywords).
    ``unit_checking=True``
        Set to ``False`` to bypass unit-checking.
    
    **Methods**
    
    .. method:: subgroup(N)
    
        Returns the next sequential subgroup of ``N`` neurons. See
        the section on subgroups below.

    .. method:: state(var)
                
        Returns the array of values for state
        variable ``var``, with length the number of neurons in the
        group.
    
    .. method:: rest()
    
        Sets the neuron state values at rest for their differential
        equations.

    The following usages are also possible for a group ``G``:
    
    ``G[i:j]``
        Returns the subgroup of neurons from ``i`` to ``j``.
    ``len(G)``
        Returns the number of neurons in ``G``.
    ``G.x``
        For any valid Python variable name ``x`` corresponding to
        a state variable of the the :class:`NeuronGroup`, this
        returns the array of values for the state
        variable ``x``, as for the :meth:`state` method
        above. Writing ``G.x = arr`` for ``arr`` a :class:`TimedArray`
        will set the values of variable x to be ``arr(t)`` at time t.
        See :class:`TimedArraySetter` for details. 
    
    **Subgroups**
    
    A subgroup is a view on a group. It isn't a new group, it's just
    a convenient way of referring to a subset of the neurons in an
    already defined group. The subset has to be a continguous set of
    neurons. They can be overlapping if defined with the slice
    notation, or consecutive if defined with the :meth:`subgroup` method.
    Subgroups can themselves be subgrouped. Subgroups can be used in
    almost all situations exactly as if they were groups, except that
    they cannot be passed to the :class:`Network` object.
    
    **Details**
    
    TODO: details of other methods and properties for people
    wanting to write extensions?
    """

    @check_units(max_delay=second)
    def __init__(self, N, model=None, threshold=None, reset=NoReset(),
                 init=None, refractory=0 * msecond, level=0,
                 clock=None, order=1, implicit=False, unit_checking=True,
                 max_delay=0 * msecond, compile=False, freeze=False, method=None,
                 max_refractory=None,
                 ):#**args): # any reason why **args was included here?
        '''
        Initializes the group.
        '''

        self._spiking = True # by default, produces spikes
        if bup.use_units: # one additional frame level induced by the decorator
            level += 1

        # If it is a string, convert to Equations object
        if isinstance(model, (str, list, tuple)):
            model = Equations(model, level=level + 1)

        if isinstance(threshold, str):
            if isinstance(model, Equations):
                threshold = select_threshold(threshold, model, level=level + 1)
            else:
                threshold = StringThreshold(threshold, level=level + 1)

        if isinstance(reset, str):
            if isinstance(model, Equations):
                reset = select_reset(reset, model, level=level + 1)
            else:
                reset = StringReset(reset, level=level + 1)

        # Clock
        clock = guess_clock(clock)#not needed with protocol checking
        self.clock = clock

        # Initial state
        self._S0 = init
        self.staticvars = []

        # StateUpdater
        if isinstance(model, StateUpdater):
            self._state_updater = model # Update mechanism
            self._all_units = defaultdict()
        elif isinstance(model, Equations):
            self._eqs = model
            if (init == None) and (model._units == {}):
                raise AttributeError, "The group must be initialized."
            self._state_updater, var_names = magic_state_updater(model, clock=clock, order=order,
                                                                 check_units=unit_checking, implicit=implicit,
                                                                 compile=compile, freeze=freeze,
                                                                 method=method)
            Group.__init__(self, model, N, unit_checking=unit_checking)
            self._all_units = model._units
            # Converts S0 from dictionary to tuple
            if self._S0 == None: # No initialization: 0 with units
                S0 = {}
            else:
                S0 = self._S0.copy()
            # Fill missing units
            for key, value in model._units.iteritems():
                if not key in S0:
                    S0[key] = 0 * value
            self._S0 = [0] * len(var_names)
            for var, i in zip(var_names, count()):
                self._S0[i] = S0[var]
        else:
            raise TypeError, "StateUpdater must be specified at initialization."
        # TODO: remove temporary unit hack, this makes all state variables dimensionless if no units are specified
        # What is this??
        if self._S0 is None:
            self._S0 = dict((i, 1.) for i in range(len(self._state_updater)))

        # Threshold
        if isinstance(threshold, Threshold):
            self._threshold = threshold
        elif type(threshold) == types.FunctionType:
            if threshold.func_code.co_argcount == 1:
                self._threshold = SimpleFunThreshold(threshold)
            else:
                self._threshold = FunThreshold(threshold)
        elif is_scalar_type(threshold):
            # Check unit
            if self._S0 != None:
                try:
                    threshold + self._S0[0]
                except DimensionMismatchError, inst:
                    raise DimensionMismatchError("The threshold does not have correct units.", *inst._dims)
            self._threshold = Threshold(threshold=threshold)
        else: # maybe raise an error?
            self._threshold = NoThreshold()
            self._spiking = False

        # Initialization of the state matrix
        if not hasattr(self, '_S'):
            self._S = zeros((len(self._state_updater), N))
        if self._S0 != None:
            for i in range(len(self._state_updater)):
                self._S[i, :] = self._S0[i]

        # Reset and refractory period
        self._variable_refractory_time = False
        period_max = 0
        if is_scalar_type(reset) or reset.__class__ is Reset:
            if reset.__class__ is Reset:
                if isinstance(reset.state, str):
                    numstate = self.get_var_index(reset.state)
                else:
                    numstate = reset.state
                reset = reset.resetvalue
            else:
                numstate = 0
            # Check unit
            if self._S0 != None:
                try:
                    reset + self._S0[numstate]
                except DimensionMismatchError, inst:
                    raise DimensionMismatchError("The reset does not have correct units.", *inst._dims)
            if isinstance(refractory, float):
                max_refractory = refractory
            else:
                if isinstance(refractory, str):
                    if max_refractory is None:
                        raise ValueError('Must specify max_refractory if using variable refractoriness.')
                    self._refractory_variable = refractory
                    self._refractory_array = None
                else:
                    max_refractory = amax(refractory) * second
                    self._refractory_variable = None
                    self._refractory_array = refractory
                self._variable_refractory_time = True
            # What is this 0.9 ?!! Answer: it's just to check that the refractory period is at least clock.dt otherwise don't bother
            if max_refractory > 0.9 * clock.dt: # Refractory period - unit checking is done here
                self._resetfun = Refractoriness(period=max_refractory, resetvalue=reset, state=numstate)
                period_max = int(max_refractory / clock.dt) + 1
            else: # Simple reset
                self._resetfun = Reset(reset, state=numstate)
        elif type(reset) == types.FunctionType:
            self._resetfun = FunReset(reset)
            if refractory > 0.9 * clock.dt:
                raise ValueError('Refractoriness for custom reset functions not yet implemented, see http://groups.google.fr/group/briansupport/browse_thread/thread/182aaf1af3499a63?hl=en for some options.')
        elif hasattr(reset, 'period'): # A reset with refractoriness
            # TODO: check unit (in Reset())
            self._resetfun = reset # reset function
            period_max = int(reset.period / clock.dt) + 1
        else: # No reset?
            self._resetfun = reset
        if hasattr(threshold, 'refractory'): # A threshold with refractoriness
            period_max = max(period_max, threshold.refractory + 1)
        if max_refractory is None:
            max_refractory = refractory
        if max_delay < period_max * clock.dt:
            max_delay = period_max * clock.dt
        self._max_delay = 0
        self.set_max_delay(max_delay)

        self._next_allowed_spiketime = -ones(N)
        self._refractory_time = float(max_refractory) - 0.5 * clock._dt
        if not self._variable_refractory_time and max_refractory < 0.9 * clock.dt:
            self._use_next_allowed_spiketime_refractoriness = False
        else:
            self._use_next_allowed_spiketime_refractoriness = True

        self._owner = self # owner (for subgroups)
        self._subgroup_set = magic.WeakSet()
        self._origin = 0 # start index from owner if subgroup
        self._next_subgroup = 0 # start index of next subgroup

        # ensure that var_index has all the 0,...,N-1 integers as names
        if not hasattr(self, 'var_index'):
            self.var_index = {}
        for i in range(self.num_states()):
            self.var_index[i] = i

        # these are here for the weave accelerated version of the threshold
        # call mechanism.
        self._spikesarray = zeros(N, dtype=int)

        # various things for optimising
        self.__t = TArray(zeros(N))
        self._var_array = {}
        for i in range(self.num_states()):
            self._var_array[i] = self._S[i]
        for kk, i in self.var_index.iteritems():
            sv = self.state_(i)
            if sv.base is self._S:
                self._var_array[kk] = sv

        # todo: should we have a guarantee that var_index exists (even if it just
        # consists of mappings i->i)?

    def set_max_delay(self, max_delay):
        if hasattr(self, '_owner') and self._owner is not self:
            self._owner.set_max_delay(max_delay)
            return
        _max_delay = int(max_delay / self.clock.dt) + 2 # in time bins
        if _max_delay > self._max_delay:
            self._max_delay = _max_delay
            self.LS = SpikeContainer(self._max_delay,
                                     useweave=get_global_preference('useweave'),
                                     compiler=get_global_preference('weavecompiler')) # Spike storage
            # update all subgroups if any exist
            if hasattr(self, '_subgroup_set'): # the first time set_max_delay is called this is false
                for G in self._owner._subgroup_set.get():
                    G._max_delay = self._max_delay
                    G.LS = self.LS

    def rest(self):
        '''
        Sets the variables at rest.
        '''
        self._state_updater.rest(self)

    def reinit(self, states=True):
        '''
        Resets the variables.
        '''
        if self._owner is self:
            if states:
                if self._S0 is not None:
                    for i in range(len(self._state_updater)):
                        self._S[i, :] = self._S0[i]
                else:
                    self._S[:] = 0 # State matrix
            self._next_allowed_spiketime[:] = -1
            self.LS.reinit()

    def update(self):
        '''
        Updates the state variables.
        '''
        self._state_updater(self) # update the variables
        if self._spiking:
            spikes = self._threshold(self) # get spikes
            if not isinstance(spikes, numpy.ndarray):
                spikes = array(spikes, dtype=int)
            if self._use_next_allowed_spiketime_refractoriness:
                spikes = spikes[self._next_allowed_spiketime[spikes] <= self.clock._t]
                if self._variable_refractory_time:
                    if self._refractory_variable is not None:
                        refractime = self.state_(self._refractory_variable)
                    else:
                        refractime = self._refractory_array
                    self._next_allowed_spiketime[spikes] = self.clock._t + refractime[spikes]
                else:
                    self._next_allowed_spiketime[spikes] = self.clock._t + self._refractory_time
            self.LS.push(spikes) # Store spikes

    def get_refractory_indices(self):
        return (self._next_allowed_spiketime > self.clock._t).nonzero()[0]

    def get_spikes(self, delay=0):
        '''
        Returns indexes of neurons that spiked at time t-delay*dt.
        '''
        if self._owner == self:
            # Group
#            if delay==0:
#                return self.LS.lastspikes()
            #return self.LS[delay] # last spikes
            return self.LS.get_spikes(delay, 0, len(self))
        else:
            # Subgroup
            return self.LS.get_spikes(delay, self._origin, len(self))
#            if delay==0:
#                ls = self.LS.lastspikes()
#            else:
#                ls = self.LS[delay]
            #ls = self.LS[delay]
#            spikes = ls-self._origin
#            return spikes[bisect.bisect_left(spikes,0):\
#                          bisect.bisect_left(spikes,len(self))]
#            return ls[bisect.bisect_left(ls,self._origin):\
#                          bisect.bisect_left(ls,len(self)+self._origin)]-self._origin

    def reset(self):
        '''
        Resets the neurons.
        '''
        self._resetfun(self)

    def subgroup(self, N):
        if self._next_subgroup + N > len(self):
            raise IndexError, "Subgroup is too large."
        P = self[self._next_subgroup:self._next_subgroup + N]
        self._next_subgroup += N;
        return P

    def unit(self, name):
        '''
        Returns the unit of variable name
        '''
        if name in self._all_units:
            return self._all_units[name]
        elif name in self.staticvars:
            f = self.staticvars[name]
            print f.func_code.co_varnames
            print [(var, self.unit(var)) for var in f.func_code.co_varnames]
            return get_unit(f(*[1. * self.unit(var) for var in f.func_code.co_varnames]))
        elif name == 't': # time
            return second
        else:
            return get_unit(self._S0[self.get_var_index(name)])

    def state_(self, name):
        if name == 't':
            self.__t[:] = self.clock._t
            return self.__t
        else:
            return Group.state_(self, name)
    state = state_

    def __getitem__(self, i):
        if i == -1:
            return self[self._S.shape[1] - 1:]
        else:
            return self[i:i + 1]

    def __getslice__(self, i, j):
        '''
        Creates subgroup (view).
        TODO: views for all arrays.
        '''
        Q = copy.copy(self)
        Q._S = self._S[:, i:j]
        Q.N = Q._S.shape[1]
        Q._origin = self._origin + i
        Q._next_subgroup = 0
        self._subgroup_set.add(Q)
        return Q

    def same(self, Q):
        '''
        Tests if the two groups (subgroups) are of the same kind,
        i.e., if they can be added.
        This is not used at the moment.
        OBSOLETE
        '''
        # Same class?
        if self.__class__ != Q.__class__:
            return False
        # Check all variables except arrays and a few ones
        exceptvar = ['owner', 'nextsubgroup', 'origin']
        for v, val in self.__dict__.iteritems():
            if not(v in Q.__dict__):
                return False
            if (not(isinstance(val, ndarray)) and (not v in exceptvar) and (val != Q.__dict__[v])):
                return False
        for v in Q.__dict__.iterkeys():
            if not(v in self.__dict__):
                return False

        return True

    def __repr__(self):
        if self._owner == self:
            return 'Group of ' + str(len(self)) + ' neurons'
        else:
            return 'Subgroup of ' + str(len(self)) + ' neurons'

    def __setattr__(self, name, val):
        global timedarray
        if timedarray is None:
            import timedarray
        if isinstance(val, timedarray.TimedArray):
            self.set_var_by_array(name, val)
        elif isinstance(val, LinkedVar):
            self.link_var(name, val.source, val.var, val.func, val.when, val.clock)
        else:
            Group.__setattr__(self, name, val)

    def link_var(self, var, source, sourcevar, func=None, when='start', clock=None):
        global network
        if network is None:
            import network
        if clock is None:
            clock = self.clock
        # check that var is not an equation (it really should only be a parameter
        # but not sure how to make this generic and still work with neurongroups
        # that aren't defined by Equations objects)
        if hasattr(self, 'staticvars') and var in self.staticvars:
            raise ValueError("Cannot set a static variable (equation) with a linked variable.")
        selfarr = self.state_(var)
        if hasattr(source, 'staticvars') and sourcevar in source.staticvars:
            if func is None: func = lambda x: x
            @network.network_operation(when=when, clock=clock)
            def update_link_var():
                selfarr[:] = func(getattr(source, sourcevar))
        else:
            sourcearr = source.state_(sourcevar)
            if func is None:
                @network.network_operation(when=when, clock=clock)
                def update_link_var():
                    selfarr[:] = sourcearr
            else:
                @network.network_operation(when=when, clock=clock)
                def update_link_var():
                    selfarr[:] = func(sourcearr)
        self._owner.contained_objects.append(update_link_var)

    def set_var_by_array(self, var, arr, times=None, clock=None, start=None, dt=None):
        # ugly hack, have to import this here because otherwise the order of imports
        # is messed up.
        import timedarray
        timedarray.set_group_var_by_array(self, var, arr, times=times, clock=clock, start=start, dt=dt)



'''
Track files that are or are not included in Brian releases.

The tracking data is stored in the folder tracking_data, and consists of
four files. The 'extras' files refer to the extras distribution and the
'release' files refer to the installation distribution. The 'rules' files give
the rules for which files should be tracked, and the 'files' files list the
files and whether or not they should be included or not. The rules files
consist of a series of regexps for which files should or shouldn't be included.
The regeps apply to the filename of the form 'brian/connection.py' (i.e. from
the brian root directory, using unix style forward slashes). The files.txt
files are a series of lines of the form 'C filename' where C is a character
which is ? if the file has not been tracked yet and requires user input, is
+ if the file should be included, - if the file shouldn't be included, and
* if the tracker should remind you whether or not it is included each time.
You can edit these files by hand.
'''
import os, sys, re, glob

pathname, filename = os.path.split(__file__)
os.chdir(pathname)
os.chdir('../../../')

#### List all the files in the Brian trunk

allfiles = []
for dirpath, dirs, files in os.walk('.'):
    if '.svn' in dirpath:
        continue
    dirs[:] = [_ for _ in dirs if '.svn' not in _]
    files = [dirpath + '/' + f for f in files]
    files = [f[2:].lower() for f in files]
    files = [f.replace('\\', '/') for f in files]
    allfiles.extend(files)

#### Extract the files that setup.py accounts for

setup = open('setup.py', 'r').read()
packages = eval(re.search(r'\bpackages\b\s*=\s*(\[.*?\])', setup, re.S).group(1).replace('\n', '').strip())
py_modules = eval(re.search(r'\bpy_modules\b\s*=\s*(\[.*?\])', setup, re.S).group(1).replace('\n', '').strip())
package_data = eval(re.search(r'\bpackage_data\b\s*=\s*(\{.*?\})', setup, re.S).group(1).replace('\n', '').strip())
extras_folders = eval(re.search(r'\bextras_folders\b\s*=\s*(\[.*?\])', setup, re.S).group(1).replace('\n', '').strip())

extras_files = []
for fold in extras_folders:
    files = glob.glob(fold)
    files = [f.replace('\\', '/').lower() for f in files]
    extras_files += files

release_files = ['readme.txt', 'setup.py']
for mod in py_modules:
    release_files.append(mod + '.py')
for p in packages:
    p = p.replace('.', '/')
    files = glob.glob(p + '/*.py')
    files = [f.replace('\\', '/').lower() for f in files]
    release_files.extend(files)
for p, pats in package_data.iteritems():
    p = p.replace('.', '/')
    for pat in pats:
        pat = p + '/' + pat
        files = glob.glob(pat)
        files = [f.replace('\\', '/').lower() for f in files]
        release_files.extend(files)

#### Load the data for given set of files

def apply_rules(rulesname, file_list):
    output_file_list = []
    for rule in open('dev/tools/newrelease/tracking_data/' + rulesname + '_rules.txt', 'r').read().split('\n'):
        if rule.startswith('include'):
            pattern = rule[8:]
            output_file_list.extend([f for f in file_list if re.search(pattern, f) and f not in output_file_list])
        elif rule.startswith('exclude'):
            pattern = rule[8:]
            output_file_list = [f for f in output_file_list if not re.search(pattern, f)]
        else:
            pass # assume it is a remark
    return output_file_list

def load_files(name, files):
    files = dict((f, '?') for f in files)
    for t in open('dev/tools/newrelease/tracking_data/' + name + '_files.txt', 'r').read().split('\n'):
        if t.strip():
            files[t[2:]] = t[0]
    # check unaccounted for files
    for f, c in files.items():
        if c == '?':
            print 'File', f, 'unaccounted for'
            print 'Choose action (+=should be included, -=should not, *=ask every time, enter to skip)'
            inp = raw_input('> ').lower().strip()
            if len(inp) == 0:
                continue
            if not inp == '+' and not inp == '*' and not inp == '-':
                print 'Bad input, skipping'
            files[f] = inp
    out = open('dev/tools/newrelease/tracking_data/' + name + '_files.txt', 'w')
    sorted_files = files.items()
    sorted_files.sort(key=lambda x:x[0])
    for f, c in sorted_files:
        print >> out, c, f
    out.close()
    return files

def get_files(name):
    files = apply_rules(name, allfiles)
    files = load_files(name, files)
    return files

#### Compare files

def compare_files(acc, files):
    ok = True
    strange = []
    too_many = []
    ask = []
    for f in acc:
        try:
            c = files[f]
            if c == '-':
                too_many.append(f)
            elif c == '*':
                ask.append(f)
            del files[f]
        except KeyError:
            if 'docs/' not in f:
                strange.append(f)
    too_few = [f for f, c in files.iteritems() if c == '+']
    ask_not_inc = [f for f, c in files.iteritems() if c == '*']
    if len(strange):
        ok = False
        print 'Files included which are not tracked (check rules):'
        print
        for f in strange:
            print f
        print
    if len(too_many):
        ok = False
        print 'Files included which should not be:'
        print
        for f in too_many:
            print f
        print
    if len(ask):
        ok = False
        print 'Files included which you have marked to be asked about each time:'
        print
        for f in ask:
            print f
        print
    if len(too_few):
        ok = False
        print 'Files not included which should be:'
        print
        for f in too_few:
            print f
        print
    if len(ask_not_inc):
        ok = False
        print 'Files not included which you have marked to be asked about each time:'
        print
        for f in ask_not_inc:
            print f
        print
    return ok

print 'RELEASE'
print '======='
print

if compare_files(release_files, get_files('release')):
    print 'OK'
    print

print 'EXTRAS'
print '======'
print
if compare_files(extras_files, get_files('extras')):
    print 'OK'
    print




from brian import *

def custom_reset(P, spikes):
    P.V_[spikes] = 0
    # plus do other stuff here...

gen = MultipleSpikeGeneratorGroup([[10 * ms, 20 * ms, 30 * ms]])
inp = NeuronGroup(1, model='V:1', threshold=0.5, reset=custom_reset)
outp = NeuronGroup(1, model='V:1', threshold=0.5, reset=0.0)
C_gen_inp = IdentityConnection(gen, inp)
C = IdentityConnection(inp, outp)
Mgen = SpikeMonitor(gen)
Minp = SpikeMonitor(inp)
Moutp = SpikeMonitor(outp)
run(100 * ms)
for M in [Mgen, Minp, Moutp]:
    print M.spikes



import pycuda.autoinit as autoinit
import pycuda.driver as drv
from pycuda.gpuarray import GPUArray
from pycuda import gpuarray
import bisect
import numpy, pylab, time
from itertools import repeat
import cProfile as profile
import pstats

N = 512 * 256
blocksize = 512
duration = 25000
record = False
doprofile = False
precision = 'double'

if precision == 'double':
    mydtype = numpy.float64
else:
    mydtype = numpy.float32

block = (blocksize, 1, 1)
grid = (N / blocksize, 1)
Ne = int(N * 0.8)
Ni = N - Ne
sparseness = 80. / N
print 'N:', N
print 'sparseness:', sparseness

mod = drv.SourceModule("""
__global__ void stateupdate(SCALAR *V, SCALAR *ge, SCALAR *gi)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x; 
    SCALAR V__tmp = (ge[i]+gi[i]-(V[i]+0.049))/0.02;
    SCALAR ge__tmp = -ge[i]/0.005;
    SCALAR gi__tmp = -gi[i]/0.01;
    V[i] += 0.0001*V__tmp;
    ge[i] += 0.0001*ge__tmp;
    gi[i] += 0.0001*gi__tmp;
}

__device__ unsigned int irng(unsigned int n)
{
    return ((n*1103515245+12345)/65536)%32768;
}
__device__ SCALAR srng(int i, int j)
{
    // formula guarantees srng(i,i)=0 which is useful for weight matrices
    return (SCALAR)(irng(irng((unsigned int)i)^irng((unsigned int)j)))/32768.0;
}
__device__ bool weightfunc(int i, int j)
{
    return srng(i,j)<SPARSENESS;
}

__global__ void threshold(SCALAR *V, int *spikes, bool *spiked, unsigned int *global_j, int N)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    bool this_spiked = V[i]>-0.05; 
    spiked[i] = this_spiked;
    if(this_spiked) // && i<N) // can leave out i<N check if N%blocksize=0
    {
        unsigned int j = atomicInc(global_j, N);
        spikes[j] = i;
    }
}

__global__ void single_thread_threshold(SCALAR *V, int *spikes, bool *spiked, unsigned int *global_j, int N)
{
    unsigned int j = 0;
    for(int i=0;i<N;i++)
    {
        if(V[i]>-0.05)
        {
            spiked[i] = true;
            spikes[j++] = i;
        } else
        {
            spiked[i] = false;
        }
    }
    *global_j = j;
}

__global__ void propagate(int *spikes, int numspikes, SCALAR *v, SCALAR W, int N, int offset)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    for(int j=0; j<numspikes; j++)
        v[i] += weightfunc(i,spikes[j]+offset)*W;
}

__global__ void reset(SCALAR *V, bool *spiked)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    V[i] = (V[i]*!spiked[i])+(-0.06)*spiked[i]; // i.e. V[i]=-0.06 if spiked[i]
}
""".replace('SCALAR', precision).replace('SPARSENESS', str(sparseness)))
stateupdate = mod.get_function("stateupdate")
threshold = mod.get_function("threshold")
single_thread_threshold = mod.get_function("single_thread_threshold")
propagate = mod.get_function("propagate")
reset = mod.get_function("reset")

V = gpuarray.to_gpu(numpy.array(numpy.random.rand(N) * 0.01 - 0.06, dtype=mydtype))
ge = gpuarray.to_gpu(numpy.zeros(N, dtype=mydtype))
gi = gpuarray.to_gpu(numpy.zeros(N, dtype=mydtype))

we = 0.00162
wi = -0.009
#CeW = numpy.array((numpy.random.rand(Ne, N)<0.02)*we,dtype=mydtype)
#CiW = numpy.array((numpy.random.rand(Ni, N)<0.02)*wi,dtype=mydtype)
#CeW = gpuarray.to_gpu(CeW)
#CiW = gpuarray.to_gpu(CiW)

gpu_spikes = drv.mem_alloc(4 * N)
gpu_spikes_exc = drv.mem_alloc(4 * N)
gpu_spikes_inh = drv.mem_alloc(4 * N)
gpu_spiked = gpuarray.to_gpu(numpy.zeros(N, dtype=bool))
gpu_spike_index = drv.mem_alloc(4)
master_spikes = numpy.zeros(N, dtype=int)
spike_index = numpy.zeros(1, dtype=numpy.uint32)
drv.memcpy_htod(gpu_spikes, master_spikes)
drv.memcpy_htod(gpu_spike_index, spike_index)

if record:
    recspikes = []
    recv0 = numpy.zeros(duration, dtype=mydtype)
totalnspikes = 0

stateupdate.prepare(('i', 'i', 'i'), block)
threshold.prepare(('i', 'i', 'i', 'i', numpy.int32), block)
single_thread_threshold.prepare(('i', 'i', 'i', 'i', numpy.int32), (1, 1, 1))
propagate.prepare(('i', numpy.int32, 'i', mydtype, numpy.int32, numpy.int32), block)
reset.prepare(('i', 'i'), block)

stateupdate_args = (grid, int(V.gpudata), int(ge.gpudata), int(gi.gpudata))
threshold_args = (grid, int(V.gpudata), int(gpu_spikes), int(gpu_spiked.gpudata), int(gpu_spike_index), numpy.int32(N))
single_thread_threshold_args = ((1, 1), int(V.gpudata), int(gpu_spikes), int(gpu_spiked.gpudata), int(gpu_spike_index), numpy.int32(N))
reset_args = (grid, int(V.gpudata), int(gpu_spiked.gpudata))

def run_sim():
    global totalnspikes
    for t in xrange(duration):
        #stateupdate(V, ge, gi, block=block, grid=grid)
        if t == 0:
            stateupdate.prepared_call(*stateupdate_args)
        else:
            stateupdate.launch_grid(*grid)
        spike_index[0] = 0
        drv.memcpy_htod(gpu_spike_index, spike_index)
        #threshold(V, gpu_spikes, gpu_spiked, gpu_spike_index, numpy.int32(N), block=block, grid=grid)
        if t == 0:
            threshold.prepared_call(*threshold_args)
        else:
            threshold.launch_grid(*grid)
        # MUCH! slower
        #single_thread_threshold.prepared_call(*single_thread_threshold_args)
        drv.memcpy_dtoh(spike_index, gpu_spike_index)
        numspikes = spike_index[0]
        spikes = master_spikes[:numspikes]
        drv.memcpy_dtoh(spikes, gpu_spikes)
        spikes.sort()
        numspikes_exc = bisect.bisect_left(spikes, Ne)
        numspikes_inh = numspikes - numspikes_exc
        drv.memcpy_htod(gpu_spikes_exc, spikes[:numspikes_exc])
        drv.memcpy_htod(gpu_spikes_inh, spikes[numspikes_exc:] - Ne)
        #propagate(gpu_spikes_exc, numpy.int32(numspikes_exc), ge, CeW, numpy.int32(N), block=block, grid=grid)
        propagate.prepared_call(grid, int(gpu_spikes_exc), numpy.int32(numspikes_exc), int(ge.gpudata), mydtype(we), numpy.int32(N), 0)
        #propagate(gpu_spikes_inh, numpy.int32(numspikes_inh), gi, CiW, numpy.int32(N), block=block, grid=grid)
        propagate.prepared_call(grid, int(gpu_spikes_inh), numpy.int32(numspikes_inh), int(gi.gpudata), mydtype(wi), numpy.int32(N), Ne)
        #reset(V, gpu_spiked, block=block, grid=grid)
        if t == 0:
            reset.prepared_call(*reset_args)
        else:
            reset.launch_grid(*grid)
        if record:
            recspikes += zip(spikes, repeat(t))
            drv.memcpy_dtoh(recv0[t:t + 1], V.gpudata)
        totalnspikes += numspikes

if doprofile:
    start = time.time()
    profile.run('run_sim()', 'cudacuba.prof')
    end = time.time()
    stats = pstats.Stats('cudacuba.prof')
    #stats.strip_dirs()
    stats.sort_stats('cumulative', 'calls')
    stats.print_stats(50)
else:
    start = time.time()
    run_sim()
    end = time.time()
print 'GPU time:', end - start
print 'Number of spikes:', totalnspikes

if record:
    i, t = zip(*recspikes)
    pylab.subplot(121)
    pylab.plot(t, i, '.')
    pylab.subplot(122)
    pylab.plot(recv0)
    pylab.show()



'''
Note: this doesn't work because you have

dw_ij/dt = c_ij(t)d(t)
c_ij(t) = sum over all pairs s, f_c(t)W(s)

where f_c(s) is an alpha function, and W(s) is the STDP learning function.

The problem is that not only is there a differential equation for w, but
there is also a differential equation for c_ij.

We could solve this by having a new system for STDP which allows you to
specify differential equations at a synaptic level.

Would even that be enough?

It would be relatively easy at least to adapt STDP so that it could do:

1. differential equations and spike time code at the separable neuronal level
2. differential equations and spike time code at the synaptic level

For example, code like:

dw/dt = -w/tau

could be turned into Python code like:

w__tmp = -w/tau
w += dt*w__tmp

where here the symbol 'w' would refer to the flattened array of all synapse
weights. Or it could be exactly solved with code like (for suitable consts):

w[:] = const0*w+const1

More synaptic variables could be introduced with extra copies of the
array. This would all work fine for sparse and dense matrices, and may be more
difficult for dynamic matrices. These would have to be linked so that shape
changes in one are mirrored in shape changes in the other. Perhaps it is worth
having a class that ensures several dynamic ConnectionMatrix objects keep the
same underlying structure of nonzero entries? It would be relatively easy to
achieve, it would just be a proxy to the other matrices.

This all works fine for zero delays, how about for heterogeneous delays?
In terms of memory use, it's fine to store a copy of the recent values of the
per-neuron variables, but unrealistic to store a copy of the recent values of
per-synapse variables in all but a very few cases. 
'''

from brian import *

sim_clock = Clock(dt=.1 * ms)
weight_update_clock = Clock(dt=10 * ms)

N = 100
tau_m = 10 * ms

G = NeuronGroup(N, 'dV/dt=-V/tau_m:1', reset=0, threshold=1, clock=sim_clock)

C = Connection(G, G, 'V', delay=True)
C_trace = Connection(G, G, 'V', delay=True)
forget(C_trace)

for i in range(N):
    C_trace[i, :] = C[i, :]
    C_trace.delay[i, :] = C.delay[i, :]

C_trace.compress()
C_trace.W.alldata[:] = 0

@network_operation(clock=weight_update_clock)
def weight_update():
    C.W.alldata += weight_update_clock.dt * C_trace.W.alldata

stdp = STDP(C_trace, '''
            dA_pre/dt  = (A_pre_aux-A_pre)/tau_pre : 1
            dA_pre_aux/dt = -A_pre_aux/tau_pre_aux : 1
            dA_post/dt = (A_post_aux-A_post)/tau_post : 1
            dA_post_aux/dt = -A_post_aux/tau_post_aux : 1
            ''', pre='''
            A_pre_aux += delta_A_pre_aux
            w += A_post
            ''', post='''
            A_post_aux += delta_A_post_aux
            w += A_pre
            ''', wmax=wmax)



import numpy
import time
import random
from itertools import repeat

# units
second = 1.
ms = 0.001
mV = 0.001

# parameters
N = 4000
Ne = int(N * 0.8)
Ni = N - Ne
Nplot = 4
dt = 0.1 * ms
T = 200 * ms
numsteps = int(T / dt)
Vr = -60 * mV
Vt = -50 * mV
p = 0.02
we = 1.62 * mV
wi = -9 * mV

# Matrices for state update, S(t+dt)=A*S+_C
A = numpy.array([[ 0.99501248, 0.00493794, 0.00496265],
 [ 0.     , 0.98019867, 0.        ],
 [ 0.  , 0.  , 0.99004983]])
_C = numpy.array([[ -2.44388520e-04],
 [ -8.58745657e-21],
 [  6.90431479e-20]])

# Initialise state matrix and assign uniform random membrane potentials
S = numpy.zeros((3, N))
S[0, :] = [random.uniform(Vr, Vt) for _ in xrange(N)]

# Generate random connectivity matrix (note: no weights)
W = []
for _ in xrange(N):
    k = numpy.random.binomial(N, p, 1)[0]
    a = random.sample(xrange(N), k)
    a.sort()
    a = numpy.array(a)
    W.append(a)

Vrec = [[] for _ in range(Nplot)]
spikesrec = []

t = 0.
Nspikes = 0
start = time.time()
for _ in xrange(numsteps):
    S[:] = numpy.dot(A, S) + _C
    spikes = (S[0, :] > Vt).nonzero()[0]
    for i in spikes:
        if i < Ne:
            S[1, W[i]] += we
        else:
            S[2, W[i]] += wi
    S[0, spikes] = Vr
    Nspikes += len(spikes)
    spikesrec += zip(spikes, repeat(t))
    t += dt
    for i in range(Nplot):
        Vrec[i].append(S[0, i])

print 'Time taken', time.time() - start
print 'Spikes', Nspikes

try:
    import pylab
    sn, st = zip(*spikesrec)
    pylab.plot(st, sn, '.')
    pylab.figure()
    for i in range(Nplot):
        pylab.plot(Vrec[i])
    pylab.show()
except ImportError:
    print 'Cannot plot, no pylab'



"""
Interplay of STDP and input oscillations
----------------------------------------
Figure 4 from:
Muller L, Brette R and Gutkin B (2011) Spike-timing dependent plasticity and 
feed-forward input oscillations produce precise and invariant spike phase-locking. Front. 
Comput. Neurosci. 5:45. doi: 10.3389/fncom.2011.00045

Description:
In this simulation, a group of IF neurons is given a tonic DC input and a tonic AC input.
The DC input is mediated by current injection (neurons.I, line 62), and the AC input is 
mediated by Poisson processes whose rate parameters are oscillating in time. Each neuron in 
the group is given a different DC input, ensuring a unique initial phase. After two seconds 
of simulation (to integrate out any initial transients), the STDP rule is turned on 
(ExponentialSTDP, line 68), and the population of neurons converges to the theoretically 
predicted fixed point. As there is some noise in the phase due to the random inputs, the 
simulation is averaged over trials (50 in Figure 4, though 10 trials should be fine for 
testing).

The trials run in parallel on all available processors (10 trials take about
2 minutes on a modern PC).
"""

### IMPORTS
from brian import *
import multiprocessing

### PARAMETERS
N=5000
M=10
taum=33*ms                          
tau_pre=20*ms
tau_post=tau_pre
Ee=0*mV
vt=-54*mV
vr=-70*mV
El=-70*mV
taue=5*ms
f=20*Hz
theta_period = 1/f
Rm=200*Mohm
a = linspace(51,65,num=M)       
weights = .001
ratio=1.50
dA_pre=.01
dA_post=.01*ratio 
trials=10

### SIMULATION LOOP
def trial(n): # n is the trial number
    reinit_default_clock()
    clear(True)

    eqs_neurons='''
    dv/dt=((ge*(Ee-vr))+Rm*I+(El-v))/taum : volt   
    dge/dt=-ge/taue : 1
    I : amp
    '''

    inputs = PoissonGroup(N,rates=lambda t:((.5-.5*cos(2*pi*f*t)))*10*Hz)           
    neurons=NeuronGroup(M,model=eqs_neurons,threshold=vt,reset=vr)
    neurons.I = a*pA
    synapses=Connection(inputs,neurons,'ge',weight=weights)
    neurons.v=vr

    S = SpikeMonitor(neurons)
    run(2*second)
    stdp=ExponentialSTDP(synapses,tau_pre,tau_post,dA_pre,-dA_post,wmax=10*weights,interactions='all',update='additive')     
    run(5*second)
    
    phase=zeros((M,200))
    for b in range(0,M):    
        tmp_phase=(S[b]%theta_period)*(360/theta_period)
        phase[b,range(0,len(tmp_phase))] = tmp_phase
        
    return phase

if __name__=='__main__': # This is very important on Windows, otherwise the machine crashes!
    phase = zeros((M,200,trials))
    
    print "This will take approximately 2 minutes."
    pool=multiprocessing.Pool() # uses all available processors b
    results=pool.map(trial,range(trials))
    for i in range(trials):
        phase[:,:,i]=results[i]
    
    ### PLOTTING
    for b in range(0,M):
        m = mean(phase[b,:,:],axis=1)
        st = std(phase[b,:,:],axis=1)/sqrt(trials)
        errorbar(range(0,135), m[range(0,135)], yerr=st[range(0,135)], xerr=None,
             fmt='-', ecolor=None, elinewidth=None, capsize=3,
             barsabove=False, lolims=False, uplims=False,
             xlolims=False, xuplims=False)
    
    title('STDP + Oscillations Simulation')
    xlabel('Spike Number')
    ylabel('Spike Phase (deg)')
    xlim([0, 135])
    ylim([140, 280])
    
    show()


from brian import *

set_global_preferences(usecstdp=False)

G = NeuronGroup(10, 'V:1')
C = Connection(G, G, 'V')

eqs_stdp = Equations('''
dx/dt = -x/second : 1
y = x*x : 1
du/dt = -u/second : 1
v = u*u : 1
''')

stdp = STDP(C, eqs=eqs_stdp, 
            pre='x+=1; w+=v',
            post='u+=1; w+=y', 
            wmax=10)



"""
SpikeQueue

This is really a sketch.
"""
#from brian.utils.circular import SpikeContainer
#from brian.neurongroup import NeuronGroup
#from brian.directcontrol import SpikeGeneratorGroup
from brian import *
from numpy import unique
from time import time

__all__=['SpikeQueue']

class SpikeQueue(NeuronGroup):
    '''
    A group that sends spikes like SpikeGeneratorGroup, but in a vectorised
    way, forgetting past events.
    Initialised with a vector of spike times and a vector of corresponding
    neuron indices.
    '''
    def __init__(self, N, spiketimes, neurons, clock=None, check_sorted=True):
        clock = guess_clock(clock)
        NeuronGroup.__init__(self, N, model=LazyStateUpdater(), clock=clock)
        # Check if spike times are sorted
        if check_sorted: # Sorts the events if necessary
            if any(diff(spiketimes)<0): # not sorted
                ind=argsort(spiketimes)
                neurons,spiketimes=neurons[ind],spiketimes[ind]
        # Create the spike queue
        self.set_max_delay(max(spiketimes)) # This leaves space for the next spikes
        # Push the spikes
        self.LS.push(neurons) # This takes a bit of time (not sure why but this could be enhanced)
        # Set the cursors back
        self.LS.ind.advance(-1)
        self.LS.S.cursor=self.LS.ind[0]
        # Discretize spike times and make them relative to current time step
        spiketimes=array((spiketimes-clock.t)/clock.dt,dtype=int) # in units of dt
        # Calculate indices of spike groups
        u,indices=unique(spiketimes,return_index=True) # determine repeated time indices
        # Build vector of indices with -1 meaning: same index as previously
        x=-ones(max(u)+2) # maximum time index
        x[-1]=len(spiketimes) # last entry
        ## This is vectorized yet incredibly inefficient 
        #x[u]=indices
        #empty_ind=where(x<0)[0] # time bins with no spikes
        # This is really slow:
        #x[empty_ind]=indices[digitize(empty_ind,u)] # -1 are replaced with next positive entry
        
        # As a loop (This now takes about 30% of the whole construction time):
        # Perhaps it could be written in C
        x[u]=indices
        for i in where(x<0)[0][::-1]: # -1 are replaced with next positive entry
            x[i]=x[i+1]
        # x[0] is always 0; maybe this should be dropped
        self.LS.ind[0:len(x)]=x
        self.LS.ind[len(x)]=-1 # no more spike at that point
        self._stopped=False # True when no more spike

    def reinit(self):
        super(SpikeQueueGroup, self).reinit()
        
    def push_spike_times(self,spiketimes,neurons):
        '''
        Inserts future spike times
        For this we need to store the end position of future spikes
        '''
        pass

    def update(self):
        # LS.S contains the data (neurons that spike)
        # LS.ind is a circular vector with pointers to locations in LS.S,
        # one for each future time bin
        if (self.LS.ind[1]>=0) & (not self._stopped):
            ns=self.LS.ind[1]-self.LS.ind[0] # number of spikes in next bin
            self.LS.S.advance(ns)
        else:
            self._stopped=True
            self.LS.ind[1]=self.LS.ind[0]
        self.LS.ind.advance(1)



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Optimizations for Brian.
'''
from scipy.weave import blitz
import re
import warnings
import parser
from inspection import *
from log import *
try:
    import sympy
    use_sympy = True
except:
    warnings.warn('sympy not installed')
    use_sympy = False
#TODO: also insert a global pref?

__all__ = ['freeze', 'simplify_expr', 'symbolic_eval']

def freeze(expr, vars, namespace={}, safe=False):
    """
    Replaces all identifiers in expr by their float value.
    The variables vars are not changed.
    If safe is True, freezing fails if one variable is not a quantity
    """
    # Find variables
    ids = [name for name in get_identifiers(expr) if name not in vars]
    # Check that they are in the namespaces and find their value
    value = {}
    for id in ids:
        if id in namespace:
            value[id] = namespace[id]
        else:
            log_warn('brian.optimizer.freeze', "Freezing impossible because the value of " + id + " is missing")
            return None
        if not isinstance(value[id], (int, float)): # or unit?
            if safe:
                log_warn('brian.optimizer.freeze', "Freezing impossible because " + id + " is not a number")
                return None
            else:
                value[id] = id
        else:
            value[id] = float(value[id]) # downcast Quantity to float
    # Substitute
    for id in ids:
        if isinstance(value[id], float):
            strver = repr(value[id])
        else:
            strver = str(value[id])
        expr = re.sub("\\b" + id + "\\b", strver, expr)
    # Clean (changes -- to +)
    expr = re.sub("--", "+", expr)
    #print "freezing:",expr
    #return simplify_expr(expr)
    return expr

def symbolic_eval(expr):
    """
    Evaluates expr as a symbolic expression.
    """
    if not use_sympy:
        return expr
    # TODO: not with all symbols
    # Find all symbols
    namespace = {}
    vars = get_identifiers(expr)
    for var in vars:
        namespace[var] = sympy.Symbol(var)
    return eval(expr, namespace)

def simplify_expr(expr):
    '''
    Simplifies a string expression for an equation.
    NB: does not seem to yield any speed up!
    '''
    return str(symbolic_eval(expr))
    #return str(sympy.simplify(symbolic_eval(expr)))



'''
In this script I test the basic Synapses behavior:
synapses = syn.Synapses(gin, gout, model = 'w : 1', pre = 'v += w')
'''

from brian import *
import dev.ideas.synapses.synapses as syn
reload(syn)
log_level_debug()

reinit_default_clock()

if False:
    # SIMPLE CASE (no delay)
    spikes = [(0, 10*ms), (1, 10*ms), (1, 20*ms), (2, 30*ms)]
    gin = SpikeGeneratorGroup(3, spikes)
    gout = NeuronGroup(2, model = 'dv/dt = -v/(5*ms) : 1')

    # init synapses
    synapses = syn.Synapses(gin, gout, model = '''w : 1; z : 1''', pre = 'v += w')
    
    # set synapses and weights
    synapses[:, 0] = True
    synapses[1, 1] = True

    synapses.w[:, 0] = 1
    synapses.w[1, 1] = 3
    
    # monitor and run
    Mv = StateMonitor(gout, 
'v', record = True)
    net = Network(gin, gout, synapses, Mv)
    net.run(100*ms)
    
    for i in range(len(gout)):
        plot(Mv.times/ms, Mv[i])
    
    show()

    
if False:
    # DELAYS
    spikes = [(0, 10*ms)]
    gin = SpikeGeneratorGroup(3, spikes)
    gout = NeuronGroup(2, model = 'dv/dt = -v/(5*ms) : 1')

    # init synapses
    synapses = syn.Synapses(gin, gout, model = '''w : 1; z : 1''', pre = 'v += w', max_delay = 3*ms)
    
    # set synapses and weights
    synapses[0, 0] = 10
    synapses.w[0, 0] = 1

    synapses.delay[:, 0] = 5*ms
    synapses.delay[:, 0, 5:] = 2*ms

    print synapses.delay

    # monitor and run
    Mv = StateMonitor(gout, 'v', record = True)
    net = Network(gin, gout, synapses, Mv)
    net.run(100*ms)
    
    for i in range(len(gout)):
        plot(Mv.times/ms, Mv[i])
    
    show()


if True:
    # Testing resizing of the spikequeue
    spikes = [(0, 10*ms)]
    
    spikes += [ (i, 30*ms) for i in range(1,10)]
    gin = SpikeGeneratorGroup(11, spikes)
    gout = NeuronGroup(2, model = 'dv/dt = -v/(5*ms) : 1')

    # init synapses
    synapses = syn.Synapses(gin, gout, model = '''w : 1; z : 1''', pre = 'v += w', max_delay = 3*ms)
    
    # set synapses and weights
    synapses[:, 0] = 10
    synapses.w[0, 0] = 1

    synapses.delay[:, 0] = 0*ms
    synapses.delay[:, 0, 5:] = 2*ms
    print synapses.delay

    # monitor and run
    Mv = StateMonitor(gout, 'v', record = True)
    net = Network(gin, gout, synapses, Mv)
    net.run(100*ms)
    
    for i in range(len(gout)):
        plot(Mv.times/ms, Mv[i])
    
    show()





from brian import *
set_global_preferences(usenewbrianhears=True)
from brian.hears import *

x = Sound(randn(15, 1), samplerate=100*Hz)
fb = DoNothingFilterbank(x)
fb.maximum_buffer_size = 5
y = fb.buffer_fetch(0, 15)

print amax(abs(y-asarray(x)))

#plot(y)
#plot(x)
#show()



from brian import *
eqs = '''
dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
dge/dt = -ge/(5*ms) : volt
dgi/dt = -gi/(10*ms) : volt
'''
P = NeuronGroup(40, model=eqs, threshold= -50 * mV, reset= -60 * mV)
P.v = -60 * mV + 10 * mV * rand(len(P))
Pe = P.subgroup(32)
Pi = P.subgroup(8)
Ce = Connection(Pe, P, 'ge')
Ci = Connection(Pi, P, 'gi')
Ce.connect_random(Pe, P, 1.0, weight=1.62 * mV)
Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)
M = SpikeMonitor(P)
run(1 * ms)

# Here do the compress and count then number of connections
Ce.compress()
countCe = 0
for i in range(0, 32):
    for j in range(0, 41):
        if abs(Ce[i, j]) > 1e-10:
            countCe += 1

print Ce.W
print countCe



from brian import *
from brian.library.synapses import *

taue = 5 * ms
gl = 20 * nS
Ee = 0 * mV
El = -60 * mV
Vt = -50 * mV
Vr = -60 * mV

eqs = MembraneEquation(C=200 * pF)
eqs += Current('I=gl*(El-vm):amp')
eqs += alpha_conductance(input='ge', E=Ee, tau=taue)

input = PoissonGroup(2, rates=50 * Hz)
G = NeuronGroup(N=1, model=eqs, threshold=Vt, reset=Vr)

C = Connection(input, G, 'ge')
C[0, 0] = 10 * nS
C[1, 0] = 5 * nS

Mvm = StateMonitor(G, 'vm', record=True)
Mge = StateMonitor(G, 'ge_current', record=True)

run(100 * ms)

figure()
subplot(211)
plot(Mvm.times, Mvm[0])
subplot(212)
plot(Mge.times, Mge[0])
show()



#!/usr/bin/env python

"""
setup.py file for SWIG example
"""

from distutils.core import setup, Extension


brianlib_module = Extension('_brianlib',
                           sources=['brianlib_wrap.cxx', 'brianlib.cpp'],
                           include_dirs=['c:\\python25\\lib\\site-packages\\numpy\\core\\include']
                           )

setup (name='brianlib',
       version='0.1',
       author="Dan Goodman",
       description="""Simple swig example from docs""",
       ext_modules=[brianlib_module],
       py_modules=["brianlib"],
       )



'''
Various base classes for Brian
'''

__all__ = ['ObjectContainer']


class ObjectContainer(object):
    '''
    Implements the contained_objects protocol
    
    The object contains an attribute _contained_objects and
    a property contained_objects whose getter just returns
    _contained_objects or an empty list, and whose setter
    appends _contained_objects with the objects. This makes
    classes which set the value of contained_objects without
    thinking about things like inheritance work correctly.
    You can still directly manipulate _contained_objects
    or do something like::
    
        co = obj.contained_objects
        co[:] = []
    '''
    def get_contained_objects(self):
        if hasattr(self, '_contained_objects'):
            return self._contained_objects
        self._contained_objects = []
        return self._contained_objects

    def set_contained_objects(self, newobjs):
        self._contained_objects = self.get_contained_objects()
        self._contained_objects.extend(newobjs)

    contained_objects = property(fget=get_contained_objects,
                                 fset=set_contained_objects)

if __name__ == '__main__':
    from brian import *

    class A(NetworkOperation):
        def __init__(self):
            x = NetworkOperation(lambda:None)
            print 'A:', id(x)
            self.contained_objects = [x]

    class B(A):
        def __init__(self):
            super(B, self).__init__()
            x = NetworkOperation(lambda:None)
            print 'B:', id(x)
            self.contained_objects = [x]

    b = B()
    print map(id, b.contained_objects)



'''
Spike-timing-dependent plasticity
'''
# See BEP-2-STDP
from inspection import *
from equations import *
from monitor import SpikeMonitor, RecentStateMonitor
from network import NetworkOperation
from neurongroup import NeuronGroup
from stateupdater import get_linear_equations, LinearStateUpdater
from scipy.linalg import expm
from scipy import dot, eye, zeros, array, clip, exp, Inf
from stdunits import ms
from connections import DelayConnection, DenseConstructionMatrix, SparseConnectionVector
import re
from utils.documentation import flattened_docstring
from copy import copy
import warnings
from itertools import izip
from numpy import arange, floor
from clock import Clock
from units import second
from utils.separate_equations import separate_equations
from log import *
from globalprefs import *
from optimiser import freeze

__all__ = ['STDP', 'ExponentialSTDP']


class STDPUpdater(SpikeMonitor):
    '''
    Updates STDP variables at spike times
    '''
    def __init__(self, source, C, vars, code, namespace, delay=0 * ms):
        '''
        source = source group
        C = connection
        vars = variable names
        M = matrix of the linear differential system
        code = code to execute for every spike
        namespace = namespace for the code
        delay = transmission delay 
        '''
        super(STDPUpdater, self).__init__(source, record=False, delay=delay)
        self._code = code # update code
        self._namespace = namespace # code namespace
        self.C = C

    def propagate(self, spikes):
        if len(spikes):
            self._namespace['spikes'] = spikes
            self._namespace['w'] = self.C.W
            exec self._code in self._namespace


class DelayedSTDPUpdater(SpikeMonitor):
    def __init__(self, C, reverse, delay_expr, max_delay,
                 vars, other_vars, varmon, othervarmon, code, namespace, delay=0 * ms):
        if reverse:
            source = C.target
            self.get_times_seq = 'get_cols'
        else:
            source = C.source
            self.get_times_seq = 'get_rows'
        super(DelayedSTDPUpdater, self).__init__(source, record=False, delay=delay)
        self._code = code # update code
        self._namespace = namespace # code namespace
        self.C = C
        self.vars = vars
        self.other_vars = other_vars
        self.varmon = varmon
        self.othervarmon = othervarmon
        delay_expr = re.sub(r'\bmax_delay\b', str(float(max_delay)), delay_expr)
        delay_expr = 'lambda d:' + delay_expr
        self.delay_expr = eval(delay_expr)

    def propagate(self, spikes):
        if len(spikes):
            if isinstance(self.get_times_seq, str):
                self.get_times_seq = getattr(self.C.delayvec, self.get_times_seq)
            times_seq = self.get_times_seq(spikes)
            times_seq = [self.delay_expr(times) for times in times_seq]
            for var in self.other_vars:
                delayed_values = self.othervarmon[var].get_past_values_sequence(times_seq)
                self._namespace[var + '__delayed_values_seq'] = delayed_values
            self._namespace['spikes'] = spikes
            self._namespace['w'] = self.C.W
            exec self._code in self._namespace


class STDP(NetworkOperation):
    '''
    Spike-timing-dependent plasticity    

    Initialised with arguments:

    ``C``
        Connection object to apply STDP to.
    ``eqs``
        Differential equations (with units)
    ``pre``
        Python code for presynaptic spikes, use the reserved symbol ``w`` to
        refer to the synaptic weight.
    ``post``
        Python code for postsynaptic spikes, use the reserved symbol ``w`` to
        refer to the synaptic weight.
    ``wmin``
        Minimum weight (default 0), weights are restricted to be within this
        value and wmax.
    ``wmax``
        Maximum weight (default unlimited), weights are restricted to be within
        wmin and this value.
    ``delay_pre``
        Presynaptic delay
    ``delay_post``
        Postsynaptic delay (backward propagating spike)
    
    The STDP object works by specifying a set of differential equations
    associated to each synapse (``eqs``) and two rules to specify what should
    happen when a presynaptic neuron fires (``pre``) and when a postsynaptic
    neuron fires (``post``). The equations should be standard set of equations
    in the usual string format. The ``pre`` and ``post`` rules should be a
    sequence of statements to be executed triggered on pre- and post-synaptic
    spikes. The sequence of statements can be separated by a ``;`` or by
    using a multiline string. The reserved symbol ``w`` can be used to refer
    to the synaptic weight of the associated synapse.
    
    This framework allows you to implement most STDP rules. Specifying
    differential equations and pre- and post-synaptic event code allows for a
    much more efficient implementation than specifying, for example, the
    spike pair weight modification function, but does unfortunately require
    transforming the definition into this form.
    
    There is one restriction on the equations that can be implemented in this
    system, they need to be separable into independent pre- and post-synaptic
    systems (this is done automatically). In this way, synaptic variables and
    updates can be stored per neuron rather than per synapse.
        
    **Example**
    
    ::
    
        eqs_stdp = """
        dA_pre/dt  = -A_pre/tau_pre   : 1
        dA_post/dt = -A_post/tau_post : 1
        """
        stdp = STDP(synapses, eqs=eqs_stdp, pre='A_pre+=delta_A_pre; w+=A_post',
                    post='A_post+=delta_A_post; w+=A_pre', wmax=gmax)

    **STDP variables**
    
    You can access the pre- and post-synaptic variables as follows::
    
        stdp = STDP(...)
        print stdp.A_pre
    
    Alternatively, you can access the group of pre/post-synaptic variables
    as::
    
        stdp.pre_group
        stdp.post_group
    
    These latter attributes can be passed to a :class:`StateMonitor` to
    record their activity, for example. However, note that in the case of
    STDP acting on a connection with heterogeneous delays, the recent values
    of these variables are automatically monitored and these can be
    accesses as follows::
    
        stdp.G_pre_monitors['A_pre']
        stdp.G_post_monitors['A_post']
    
    **Technical details**
    
    The equations are split into two groups, pre and post. Two groups are created
    to carry these variables and to update them (these are implemented as
    :class:`NeuronGroup` objects). As well as propagating spikes from the source
    and target of ``C`` via ``C``, spikes are also propagated to the respective
    groups created. At spike propagation time the weight values are updated.
    '''
    def __init__(self, C, eqs, pre, post, wmin=0, wmax=Inf, level=0, clock=None, delay_pre=None, delay_post=None):
        '''
        C: connection object
        eqs: differential equations (with units)
        pre: Python code for presynaptic spikes
        post: Python code for postsynaptic spikes
        wmax: maximum weight (default unlimited)
        delay_pre: presynaptic delay
        delay_post: postsynaptic delay (backward propagating spike)
        '''
        if get_global_preference('usecstdp') and get_global_preference('useweave'):
            from experimental.c_stdp import CSTDP
            log_warn('brian.stdp', 'Using experimental C STDP class.')
            self.__class__ = CSTDP
            CSTDP.__init__(self, C, eqs, pre, post, wmin=wmin, wmax=wmax,
                           level=level + 1, clock=clock, delay_pre=delay_pre,
                           delay_post=delay_post)
            return
        NetworkOperation.__init__(self, lambda:None, clock=clock)
        # Convert to equations object
        if isinstance(eqs, Equations):
            eqs_obj = eqs
        else:
            eqs_obj = Equations(eqs, level=level + 1)
        # handle multi-line pre, post equations and multi-statement equations separated by ;
        if '\n' in pre:
            pre = flattened_docstring(pre)
        elif ';' in pre:
            pre = '\n'.join([line.strip() for line in pre.split(';')])
        if '\n' in post:
            post = flattened_docstring(post)
        elif ';' in post:
            post = '\n'.join([line.strip() for line in post.split(';')])

        # Check units
        eqs_obj.compile_functions()
        eqs_obj.check_units()
        # Get variable names
        vars = eqs_obj._diffeq_names
        # Find which ones are directly modified (e.g. regular expression matching; careful with comments)
        vars_pre = [var for var in vars if var in modified_variables(pre)]
        vars_post = [var for var in vars if var in modified_variables(post)]

        # additional dependencies are used to ensure that if there are multiple
        # pre/post separated equations they are grouped together as one
        #
        # We should replace with this code here, in case there are no pre/post synaptic variables
        #additional_deps =[]
        #if len(vars_pre)>0:
        #    additional_deps.append('__pre_deps='+'+'.join(vars_pre))
        #if len(vars_post)>0:
        #    additional_deps.append('__post_deps='+'+'.join(vars_post))
        additional_deps = ['__pre_deps='+'+'.join(vars_pre),
                           '__post_deps='+'+'.join(vars_post)]
        separated_equations = separate_equations(eqs_obj, additional_deps)
        if not len(separated_equations) == 2:
            raise ValueError('Equations should separate into pre and postsynaptic variables.')
        sep_pre, sep_post = separated_equations
        for v in vars_pre:
            if v in sep_post._diffeq_names:
                sep_pre, sep_post = sep_post, sep_pre
                break
        index_pre = [i for i in range(len(vars)) if vars[i] in vars_pre or vars[i] in sep_pre._diffeq_names]
        index_post = [i for i in range(len(vars)) if vars[i] in vars_post or vars[i] in sep_post._diffeq_names]

        vars_pre = array(vars)[index_pre]
        vars_post = array(vars)[index_post]

        # Check pre/post consistency
        shared_vars = set(vars_pre).intersection(vars_post)
        if shared_vars != set([]):
            raise Exception, str(list(shared_vars)) + " are both presynaptic and postsynaptic!"

        # Substitute equations/aliases into pre/post code
        def substitute_eqs(code):
            for name in sep_pre._eq_names[-1::-1]+sep_post._eq_names[-1::-1]: # reverse order, as in equations.py
                if name in sep_pre._eq_names:
                    expr = sep_pre._string[name]
                else:
                    expr = sep_post._string[name]
                code = re.sub("\\b" + name + "\\b", '(' + expr + ')', code)
            return code
        pre = substitute_eqs(pre)
        post = substitute_eqs(post)

        # Create namespaces for pre and post codes
        pre_namespace = namespace(pre, level=level + 1)
        post_namespace = namespace(post, level=level + 1)
        pre_namespace['clip'] = clip
        post_namespace['clip'] = clip
        pre_namespace['Inf'] = Inf
        post_namespace['Inf'] = Inf
        pre_namespace['enumerate'] = enumerate
        post_namespace['enumerate'] = enumerate

        # freeze pre and post (otherwise units will cause problems)
        all_vars = list(vars_pre) + list(vars_post) + ['w']
        pre = '\n'.join(freeze(line.strip(), all_vars, pre_namespace) for line in pre.split('\n'))
        post = '\n'.join(freeze(line.strip(), all_vars, post_namespace) for line in post.split('\n'))

        # Neuron groups
        G_pre = NeuronGroup(len(C.source), model=sep_pre, clock=self.clock)
        G_post = NeuronGroup(len(C.target), model=sep_post, clock=self.clock)
        G_pre._S[:] = 0
        G_post._S[:] = 0
        self.pre_group = G_pre
        self.post_group = G_post
        var_group = {} # maps variable name to group
        for v in vars_pre:
            var_group[v] = G_pre
        for v in vars_post:
            var_group[v] = G_post
        self.var_group = var_group

        # Create updaters and monitors
        if isinstance(C, DelayConnection):
            G_pre_monitors = {} # these get values put in them later
            G_post_monitors = {}
            max_delay = C._max_delay * C.target.clock.dt

            def gencode(incode, vars, other_vars, wreplacement):
                num_immediate = num_delayed = 0
                reordering_warning = False
                incode_lines = [line.strip() for line in incode.split('\n')]
                outcode_immediate = 'for _i in spikes:\n'
                # delayed variables
                outcode_delayed = 'for _j, _i in enumerate(spikes):\n'
                for var in other_vars:
                    outcode_delayed += '    ' + var + '__delayed = ' + var + '__delayed_values_seq[_j]\n'
                for line in incode_lines:
                    if not line.strip(): continue
                    m = re.search(r'\bw\b\s*[^><=]?=', line) # lines of the form w = ..., w *= ..., etc.
                    for var in vars:
                        line = re.sub(r'\b' + var + r'\b', var + '[_i]', line)
                    for var in other_vars:
                        line = re.sub(r'\b' + var + r'\b', var + '__delayed', line)
                    if m:
                        num_delayed += 1
                        outcode_delayed += '    ' + line + '\n'
                    else:
                        if num_delayed!=0 and not reordering_warning:
                            log_warn('brian.stdp', 'STDP operations are being re-ordered for delay connection, results may be wrong.')
                            reordering_warning = True
                        num_immediate += 1
                        outcode_immediate += '    ' + line + '\n'
                outcode_delayed = re.sub(r'\bw\b', wreplacement, outcode_delayed)
                outcode_delayed += '\n    %(w)s = clip(%(w)s, %(min)e, %(max)e)' % {'min':wmin, 'max':wmax, 'w':wreplacement}
                return (outcode_immediate, outcode_delayed)

            pre_immediate, pre_delayed = gencode(pre, vars_pre, vars_post, 'w[_i,:]')
            post_immediate, post_delayed = gencode(post, vars_post, vars_pre, 'w[:,_i]')
            log_debug('brian.stdp', 'PRE CODE IMMEDIATE:\n'+pre_immediate)
            log_debug('brian.stdp', 'PRE CODE DELAYED:\n'+pre_delayed)
            log_debug('brian.stdp', 'POST CODE:\n'+post_immediate+post_delayed)
            pre_delay_expr = 'max_delay-d'
            post_delay_expr = 'd'
            pre_code_immediate = compile(pre_immediate, "Presynaptic code immediate", "exec")
            pre_code_delayed = compile(pre_delayed, "Presynaptic code delayed", "exec")
            post_code = compile(post_immediate + post_delayed, "Postsynaptic code", "exec")

            if delay_pre is not None or delay_post is not None:
                raise ValueError("Must use delay_pre=delay_post=None for the moment.")
            max_delay = C._max_delay * C.target.clock.dt
            # Ensure that the source and target neuron spikes are kept for at least the
            # DelayConnection's maximum delay
            C.source.set_max_delay(max_delay)
            C.target.set_max_delay(max_delay)
            # create forward and backward Connection objects or SpikeMonitor objects
            pre_updater_immediate = STDPUpdater(C.source, C, vars=vars_pre,
                                           code=pre_code_immediate, namespace=pre_namespace, delay=0 * ms)
            pre_updater_delayed = DelayedSTDPUpdater(C, reverse=False, delay_expr=pre_delay_expr, max_delay=max_delay,
                                            vars=vars_pre, other_vars=vars_post,
                                            varmon=G_pre_monitors, othervarmon=G_post_monitors,
                                            code=pre_code_delayed, namespace=pre_namespace, delay=max_delay)
            post_updater = DelayedSTDPUpdater(C, reverse=True, delay_expr=post_delay_expr, max_delay=max_delay,
                                            vars=vars_post, other_vars=vars_pre,
                                            varmon=G_post_monitors, othervarmon=G_pre_monitors,
                                            code=post_code, namespace=post_namespace, delay=0 * ms)
            updaters = [pre_updater_immediate, pre_updater_delayed, post_updater]
            self.contained_objects += updaters
            vars_pre_ind = dict((var, i) for i, var in enumerate(vars_pre))
            vars_post_ind = dict((var, i) for i, var in enumerate(vars_post))
            self.G_pre_monitors = G_pre_monitors
            self.G_post_monitors = G_post_monitors
            self.G_pre_monitors.update(((var, RecentStateMonitor(G_pre, vars_pre_ind[var], duration=(C._max_delay + 1) * C.target.clock.dt, clock=G_pre.clock)) for var in vars_pre))
            self.G_post_monitors.update(((var, RecentStateMonitor(G_post, vars_post_ind[var], duration=(C._max_delay + 1) * C.target.clock.dt, clock=G_post.clock)) for var in vars_post))
            self.contained_objects += self.G_pre_monitors.values()
            self.contained_objects += self.G_post_monitors.values()

        else:
            # Indent and loop
            pre = re.compile('^', re.M).sub('    ', pre)
            post = re.compile('^', re.M).sub('    ', post)
            pre = 'for _i in spikes:\n' + pre
            post = 'for _i in spikes:\n' + post

            # Pre code
            for var in vars_pre: # presynaptic variables (vectorisation)
                pre = re.sub(r'\b' + var + r'\b', var + '[_i]', pre)
            pre = re.sub(r'\bw\b', 'w[_i,:]', pre) # synaptic weight
            # Post code
            for var in vars_post: # postsynaptic variables (vectorisation)
                post = re.sub(r'\b' + var + r'\b', var + '[_i]', post)
            post = re.sub(r'\bw\b', 'w[:,_i]', post) # synaptic weight

            # Bounds: add one line to pre/post code (clip(w,min,max,w))
            # or actual code? (rather than compiled string)
            if wmax==Inf:
                pre += '\n    w[_i,:]=clip(w[_i,:],%(min)e,Inf)' % {'min':wmin}
                post += '\n    w[:,_i]=clip(w[:,_i],%(min)e,Inf)' % {'min':wmin}
            else:
                pre += '\n    w[_i,:]=clip(w[_i,:],%(min)e,%(max)e)' % {'min':wmin, 'max':wmax}
                post += '\n    w[:,_i]=clip(w[:,_i],%(min)e,%(max)e)' % {'min':wmin, 'max':wmax}
            log_debug('brian.stdp', 'PRE CODE:\n'+pre)
            log_debug('brian.stdp', 'POST CODE:\n'+post)
            # Compile code
            pre_code = compile(pre, "Presynaptic code", "exec")
            post_code = compile(post, "Postsynaptic code", "exec")

            connection_delay = C.delay * C.source.clock.dt
            if (delay_pre is None) and (delay_post is None): # same delays as the Connnection C
                delay_pre = connection_delay
                delay_post = 0 * ms
            elif delay_pre is None:
                delay_pre = connection_delay - delay_post
                if delay_pre < 0 * ms: raise AttributeError, "Presynaptic delay is too large"
            elif delay_post is None:
                delay_post = connection_delay - delay_pre
                if delay_post < 0 * ms: raise AttributeError, "Postsynaptic delay is too large"
            # create forward and backward Connection objects or SpikeMonitor objects
            pre_updater = STDPUpdater(C.source, C, vars=vars_pre, code=pre_code, namespace=pre_namespace, delay=delay_pre)
            post_updater = STDPUpdater(C.target, C, vars=vars_post, code=post_code, namespace=post_namespace, delay=delay_post)
            updaters = [pre_updater, post_updater]
            self.contained_objects += [pre_updater, post_updater]

        # Put variables in namespaces
        for i, var in enumerate(vars_pre):
            for updater in updaters:
                updater._namespace[var] = G_pre._S[i]
        for i, var in enumerate(vars_post):
            for updater in updaters:
                updater._namespace[var] = G_post._S[i]

        self.contained_objects += [G_pre, G_post]

    def __call__(self):
        pass

    def __getattr__(self, name):
        if name == 'var_group':
            # this seems mad - the reason is that getattr is only called if the thing hasn't
            # been found using the standard methods of finding attributes, which for var_index
            # should have worked, this is important because the next line looks for var_index
            # and if we haven't got a var_index we don't want to get stuck in an infinite
            # loop
            raise AttributeError
        if not hasattr(self, 'var_group'):
            # only provide lookup of variable names if we have some variable names, i.e.
            # if the var_index attribute exists
            raise AttributeError
        G = self.var_group[name]
        return G.state_(name)

    def __setattr__(self, name, val):
        if not hasattr(self, 'var_group') or name not in self.var_group:
            object.__setattr__(self, name, val)
        else:
            G = self.var_group[name]
            G.state_(name)[:] = val


class ExponentialSTDP(STDP):
    '''
    Exponential STDP.
    
    Initialised with the following arguments:
    
    ``taup``, ``taum``, ``Ap``, ``Am``
        Synaptic weight change (relative to the maximum weight wmax)::
        
            f(s) = Ap*exp(-s/taup) if s >0
            f(s) = Am*exp(s/taum) if s <0
                    
    ``interactions``
      * 'all': contributions from all pre-post pairs are added
      * 'nearest': only nearest-neighbour pairs are considered
      * 'nearest_pre': nearest presynaptic spike, all postsynaptic spikes
      * 'nearest_post': nearest postsynaptic spike, all presynaptic spikes
    
    ``wmin=0``
        minimum synaptic weight
        
    ``wmax``
        maximum synaptic weight
        
    ``update``
      * 'additive': modifications are additive (independent of synaptic weight)
        (or "hard bounds")
      * 'multiplicative': modifications are multiplicative (proportional to w)
        (or "soft bounds")
      * 'mixed': depression is multiplicative, potentiation is additive
    
    See documentation for :class:`STDP` for more details.
    '''
    def __init__(self, C, taup, taum, Ap, Am, interactions='all', wmin=0, wmax=None,
                 update='additive', delay_pre=None, delay_post=None, clock=None):
        if wmax is None:
            raise AttributeError, "You must specify the maximum synaptic weight"
        wmax = float(wmax) # removes units

        eqs = Equations('''
        dA_pre/dt=-A_pre/taup : 1
        dA_post/dt=-A_post/taum : 1''', taup=taup, taum=taum, wmax=wmax)
        if interactions == 'all':
            pre = 'A_pre+=Ap'
            post = 'A_post+=Am'
        elif interactions == 'nearest':
            pre = 'A_pre=Ap'
            post = 'A_post=Am'
        elif interactions == 'nearest_pre':
            pre = 'A_pre=Ap'
            post = 'A_post=+Am'
        elif interactions == 'nearest_post':
            pre = 'A_pre+=Ap'
            post = 'A_post=Am'
        else:
            raise AttributeError, "Unknown interaction type " + interactions

        if update == 'additive':
            Ap *= wmax
            Am *= wmax
            pre += '\nw+=A_post'
            post += '\nw+=A_pre'
        elif update == 'multiplicative':
            if Am < 0:
                pre += '\nw*=(1+A_post)'
            else:
                pre += '\nw+=(wmax-w)*A_post'
            if Ap < 0:
                post += '\nw*=(1+A_pre)'
            else:
                post += '\nw+=(wmax-w)*A_pre'
        elif update == 'mixed':
            if Am < 0 and Ap > 0:
                Ap *= wmax
                pre += '\nw*=(1+A_post)'
                post += '\nw+=A_pre'
            elif Am > 0 and Ap < 0:
                Am *= wmax
                post += '\nw*=(1+A_pre)'
                pre += '\nw+=A_post'
            else:
                if Am > 0:
                    raise AttributeError, "There is no depression in STDP rule"
                else:
                    raise AttributeError, "There is no potentiation in STDP rule"
        else:
            raise AttributeError, "Unknown update type " + update
        STDP.__init__(self, C, eqs=eqs, pre=pre, post=post, wmin=wmin, wmax=wmax, delay_pre=delay_pre, delay_post=delay_post, clock=clock)

if __name__ == '__main__':
    pass



'''
This example implements the artificial vowels from
Culling, J. F. and Summerfield, Q. (1995a). "Perceptual segregation of
concurrent speech sounds: absence of across-frequency grouping by common
interaural delay" J. Acoust. Soc. Am. 98, 785-797.
'''
from brian import *
from brian.hears import *

duration = 409.6*ms
width = 150*Hz/2
samplerate = 10*kHz

set_default_samplerate(samplerate)

centres = [225*Hz, 625*Hz, 975*Hz, 1925*Hz]
vowels = {
    'ee':[centres[0], centres[3]],
    'ar':[centres[1], centres[2]],
    'oo':[centres[0], centres[2]],
    'er':[centres[1], centres[3]]
    }

def generate_vowel(vowel):
    vowel = vowels[vowel]
    x = whitenoise(duration)
    y = fft(asarray(x).flatten())
    f = fftfreq(len(x), 1/samplerate)
    I = zeros(len(f), dtype=bool)
    for cf in vowel:
        I = I|((abs(f)<cf+width)&(abs(f)>cf-width))
    I = -I
    y[I] = 0
    x = ifft(y)
    return Sound(x.real)

v1 = generate_vowel('ee').ramp()
v2 = generate_vowel('ar').ramp()
v3 = generate_vowel('oo').ramp()
v4 = generate_vowel('er').ramp()

for s in [v1, v2, v3, v4]:
    s.play(normalise=True, sleep=True)

s1 = Sound((v1, v2))
#s1.play(normalise=True, sleep=True)

s2 = Sound((v3, v4))
#s2.play(normalise=True, sleep=True)

v1.save('mono_sound.wav')
s1.save('stereo_sound.wav')

subplot(211)
plot(v1.times, v1)
subplot(212)
v1.spectrogram()
show()



from brian import *
import dev.ideas.synapses.synapses as syn
reload(syn)

existing_synapses = np.array([[1,2,3], [4,5,6]])
groups_shape = (4, 7)
data = np.zeros(existing_synapses.shape[1])

pv = syn.ParameterVector(data, groups_shape, existing_synapses)


print 'initial', pv.data
pv[1,:] = 3


print 'modified', pv.data
print data




'''
Tutorial 1e: Connecting neurons
*******************************

In the previous parts of this tutorial, the neurons are
still all unconnected. We add in connections here. The
model we use is that when neuron i is connected to 
neuron j and neuron i fires a spike, then the membrane
potential of neuron j is instantaneously increased by
a value ``psp``. We start as before:
'''
from brian import *

tau = 20 * msecond        # membrane time constant
Vt = -50 * mvolt          # spike threshold
Vr = -60 * mvolt          # reset value
El = -49 * mvolt          # resting potential (same as the reset)
'''
Now we include a new parameter, the PSP size:
'''
psp = 0.5 * mvolt         # postsynaptic potential size
'''
And continue as before:
'''
G = NeuronGroup(N=40, model='dV/dt = -(V-El)/tau : volt',
              threshold=Vt, reset=Vr)
'''
Connections
~~~~~~~~~~~

We now proceed to connect these neurons. Firstly, we declare
that there is a connection from neurons in ``G`` to neurons in ``G``.
For the moment, this is just something that is necessary to
do, the reason for doing it this way will become clear in the
next tutorial.
'''
C = Connection(G, G)
'''
Now the interesting part, we make these neurons be randomly
connected with probability 0.1 and weight ``psp``. Each neuron
i in ``G`` will be connected to each neuron j in ``G`` 
with probability 0.1. The weight of the connection is the
amount that is added to the membrane potential of the target
neuron when the source neuron fires a spike.
'''
C.connect_random(sparseness=0.1, weight=psp)
'''
These two previous lines could be done in one line::

  C = Connection(G,G,sparseness=0.1,weight=psp)

Now we continue as before:
'''
M = SpikeMonitor(G)

G.V = Vr + rand(40) * (Vt - Vr)

run(1 * second)

print M.nspikes
'''
You can see that the number of spikes has jumped from around
800-850 to around 1000-1200. In the next part of the tutorial,
we'll look at a way to plot the output of the network.

Exercise
~~~~~~~~

Try varying the parameter ``psp`` and see what happens. How large
can you make the number of spikes output by the network? Why?

Solution
~~~~~~~~

The logically maximum number of firings is
400,000 = 40 * 1000 / 0.1, the number of neurons in the
network * the time it runs for / the integration step size (you
cannot have more than one spike per step).

In fact, the number of firings is bounded above by 200,000. The
reason for this is that the network updates in the following way:

1. Integration step
2. Find neurons above threshold
3. Propagate spikes
4. Reset neurons which spiked

You can see then that if neuron i has spiked at time t, then it
will not spike at time t+dt, even if it receives spikes from
another neuron. Those spikes it receives will be added at step
3 at time t, then reset to ``Vr`` at step 4 of time t, then the
thresholding function at time t+dt is applied at step 2, before
it has received any subsequent inputs. So the most a neuron
can spike is every other time step.
'''



'''
Conversion from integral form to differential form.
See BEP-5.
 
TODO:
* maximum rank
* better function name
* discrete time version
* rescale X0 to avoid numerical problems
* automatic determination of T?
'''
import re
import inspect
from brian.units import *
from brian.stdunits import *
from brian.inspection import get_identifiers
from brian.utils.autodiff import *
from brian.equations import *
from scipy import linalg

def integral2differential(expr, T=20 * ms, level=0, N=20, suffix=None, matrix_output=False):
    '''
    Example:
      eqs,w=integral2differential('g(t)=t*exp(-t/tau)')
      M,nvar,w=integral2differential('g(t)=t*exp(-t/tau)',matrix_output=True)
      
    Returns an Equations object corresponding to the time-invariant linear system specified
    by the impulse response g(t), and the value w to generate the impulse response:
    g_in->g_in+w.
    
    If matrix_output is True, returns the matrix of the corresponding differential system, the
    index nvar of the variable and the initial condition w=x_nvar(0).
    
    T is the interval over which the function is calculated.
    N is the number of points chosen in that interval.
    level is the frame level where the expression is defined.
    suffix is a string added to internal variable names (default: unique string).
    '''
    # Expression matching
    varname, time, RHS = re.search('\s*(\w+)\s*\(\s*(\w+)\s*\)\s*=\s*(.+)\s*', expr).groups()

    # Build the namespace
    frame = inspect.stack()[level + 1][0]
    global_namespace, local_namespace = frame.f_globals, frame.f_locals
    # Find external objects
    vars = list(get_identifiers(RHS))
    namespace = {}
    for var in vars:
        if var == time: # time variable
            pass
        elif var in local_namespace: #local
            namespace[var] = local_namespace[var]
        elif var in global_namespace: #global
            namespace[var] = global_namespace[var]
        elif var in globals(): # typically units
            namespace[var] = globals()[var]

    # Convert to a function
    f = eval('lambda ' + time + ':' + RHS, namespace)

    # Unit
    unit = get_unit(f(rand()*second)).name

    # Pick N points
    t = rand(N) * T

    # Calculate derivatives and find rank
    n = 0
    rank = 0
    M = f(t).reshape(N, 1)
    while rank == n:
        n += 1
        dfn = differentiate(f, t, order=n).reshape(N, 1)
        x, _, rank, _ = linalg.lstsq(M, dfn)
        if rank == n:
            M = hstack([M, dfn])
            oldx = x
    # oldx expresses dfn as a function of df0,..,dfn-1 (n=rank)

    # Find initial condition
    X0 = array([differentiate(f, 0 * ms, order=n) for n in range(rank)])

    # Rescaling DOES NOT WORK
    #R=ones(rank)
    #for i in range(rank):
    #    if X0[i]!=0.:
    #        R[i]=1./X0[i]
    #    else:
    #        R[i]=1.
    #R=diag(R)
    #X0=dot(R,X0)
    #oldx=dot(R,oldx)

    # Build A
    A = diag(ones(rank - 1), 1)
    A[-1, :] = oldx.reshape(1, rank)

    # Find Q=P^{-1}
    Q = eye(rank)
    if X0[0] == 0.: # continuous g, spikes act on last variable: x->x+1
        Q[:, -1] = X0
        nvar = rank - 1
        w = 1.
        # Exact inversion
        P = eye(rank)
        P[:-1, -1] = -X0[:-1] / X0[-1] # Has to be !=0 !!
        P[-1, -1] = 1. / X0[-1]
    else: # discontinuous g, spikes act on first variable: x->x+g(0)
        Q[:, 0] = X0
        nvar = 0
        w = X0[0]
        P = linalg.inv(Q)

    M = dot(dot(P, A), Q)
    #M=dot(linalg.inv(R),dot(M,R))

    # Turn into string
    # Set variable names
    if rank < 5:
        names = [varname] + ['x', 'y', 'z'][:rank - 1]
    else:
        names = [varname] + ['x' + str(i) for i in range(rank - 1)]

    # Add suffix
    if suffix is None:
        suffix = unique_id()
    names[1:] = [name + suffix for name in names[1:]]

    # Build string
    eqs = []
    for i in range(rank):
        eqs.append('d' + names[i] + '/dt=' + '+'.join([str(x) + '*' + name for x, name in zip(M[i, :], names) if x != 0.]) +
                   ' : ' + str(unit))
    eqs.append(varname + '_in=' + names[nvar]) # alias
    eq_string = '\n'.join(eqs).replace('+-', '-')

    if matrix_output:
        return M, nvar, w
    else:
        return Equations(eq_string), w

if __name__ == '__main__':
    from brian import *
    from scipy import linalg
    tau = 10 * ms
    tau2 = 5 * ms
    freq = 350 * Hz
    # The gammatone example does not seem to work for higher orders
    # probably a numerical problem; use a rescaling matrix for X0?
    f = lambda t:(t / tau) ** 1 * exp(-t / tau) * cos(2 * pi * freq * t)
    A, nvar, w = integral2differential('g(t)=(t/tau)**1*exp(-t/tau)*cos(2*pi*freq*t)', suffix='',
                                   matrix_output=True)
    eq, w = integral2differential('g(t)=(t/tau)**1*exp(-t/tau)*cos(2*pi*freq*t)', suffix='',
                                   matrix_output=False)
    print eq
    #f=lambda t:exp(-t/tau)-exp(-t/tau2)*cos(2*pi*t/tau)
    #A,nvar,w=integral2differential('g(t)=exp(-t/tau)-exp(-t/tau2)*cos(2*pi*t/tau)',
    #                                matrix_output=True)
    print A, nvar, w
    for t in range(10):
        t = t * 1 * ms
        print linalg.expm(A * t)[0, nvar] * w, f(t)
    #t=arange(50)*.5*ms
    #plot(t,f(t))
    #show()



import brian_no_units
from brian import *
set_global_preferences(useweave=True)
import cuba
from cuba_runopts import N_varywe
cuba.do_runs_varywe('data/brian_cuba_varywe_results_compile_nounits_' + str(N_varywe) + '.pkl')



'''
Voltage-clamp experiment (SEVC)
'''
from brian import *
from brian.library.electrophysiology import *

defaultclock.dt = .01 * ms

taum = 20 * ms        # membrane time constant
gl = 1. / (50 * Mohm)   # leak conductance
Cm = taum * gl        # membrane capacitance
Re = 50 * Mohm        # electrode resistance
Ce = 0.1 * ms / Re      # electrode capacitance

eqs = Equations('''
dvm/dt=(-gl*vm+i_inj)/Cm : volt
I:amp
''')
eqs += current_clamp(i_cmd='I', Re=Re, Ce=Ce)
setup = NeuronGroup(1, model=eqs)
ampli = SEVC(setup, 'v_rec', 'I', 1 * kHz, gain=250 * nS, gain2=50 * nS / ms)
recording = StateMonitor(ampli, 'record', record=True)
soma = StateMonitor(setup, 'vm', record=True)

ampli.command = 20 * mV
run(200 * ms)
figure()
plot(recording.times / ms, recording[0] / nA, 'k')
figure()
plot(soma.times / ms, soma[0] / mV, 'b')
show()



from brian import *

from brian.units import check_units, siemens, volt
from brian.membrane_equations import Current

def leak_current_int(gl, El, current_name=None):
     return Current('I=gl*(El-vm) : amp', gl=gl, El=El, I=current_name)

def K_current_int(gmax, EK, current_name=None):
     return Current('''
     I=gmax*n**4*(EK-vm) : amp
     dn/dt=alphan*(1-n)-betan*n : 1
     alphan=.05*(vm+34.*mV)/(1-exp(-(vm+34.*mV)/(10.*mV)))/mV/ms : Hz
     betan=.625*exp(-(vm+44.*mV)/(80.*mV))/ms : Hz
     ''', gmax=gmax, EK=EK, I=current_name)

def Na_current_int(gmax, ENa, current_name=None):
     return Current('''
     I=gmax/((1+betam/alpham)**3)*h*(ENa-vm) : amp
     dh/dt=alphah*(1-h)-betah*h : 1
     alpham=.5*(vm+35.*mV)/(1-exp(-(vm+35.*mV)/(10.*mV)))/mV/ms : Hz
     betam=20.*exp(-(vm+60.*mV)/(18.*mV))/ms : Hz
     alphah=.35*exp(-(vm+58.*mV)/(20.*mV))/ms : Hz
     betah=5./(1.+exp(-(vm+28.*mV)/(10.*mV)))/ms : Hz
     ''', gmax=gmax, ENa=ENa, I=current_name)

c = Clock(dt=.01 * ms) # more precise
El = -65 * mV
EK = -90 * mV
ENa = 55 * mV
eqs = MembraneEquation(0.01 * uF) + leak_current_int(.003 * msiemens, El, current_name='Il')
eqs += K_current_int(0.36 * msiemens, EK, current_name='IK')
eqs += Na_current_int(1.20 * msiemens, ENa)
eqs += Current(I='Iapp:amp')

neuron = NeuronGroup(1, eqs)

neuron.h = 1.
neuron.n = .1
neuron.vm = -64 * mV

trace = StateMonitor(neuron, 'vm', record=True)

run(100 * ms)
neuron.Iapp = 0.

run(100 * ms)
print trace[0] / mV
plot(trace.times / ms, trace[0] / mV)
show()



from numpy import *
import time

spikes = arange(10000) # presynaptic neurons that spiked
pre = ones(100) # presyn neuron per synapse
pre[50:] = arange(50)
pre[0] = 1


'''
Spec:

function(spikes, pre) should return the array of synapse numbers concerned by the spikes in spikes (a list)

Issues:
- multiple synapse per same presyn neuron
- speed
'''

def dummy_method(s, p):
    synapses = []
    for i in s:
        idx = list(nonzero(p == i)[0])
        synapses+=idx
    return synapses

if True:
    t0 = time.time()
    resdummy = dummy_method(spikes, pre)
    print 'dummy method in ', time.time()-t0


def meshgrid_method(s, p):
    ss, pp = meshgrid(s, p)
    synapses = nonzero(sum(ss == pp, axis = 1))[0]
    return synapses

if True:
    t0 = time.time()
    resmesh = meshgrid_method(spikes, pre)
    print 'mesh method in ', time.time()-t0

print (sort(resmesh) == sort(resdummy)).all()




from brian import *
from nose.tools import *
import numpy

def test():
    """
    Names
    ~~~~~
    
    The following units should exist:
    
        metre, kilogram, second, amp, kelvin, mole, candle
        radian, steradian, hertz, newton, pascal, joule, watt,
        coulomb, volt, farad, ohm, siemens, weber, tesla, henry,
        celsius, lumen, lux, becquerel, gray, sievert, katal,
        gram, gramme
    
    In addition, all versions of these units scaled by the following
    prefixes should exist (in descending order of size):
    
        Y, Z, E, P, T, G, M, k, h, da, d, c, m, u, n, p, f, a, z, y
    
    And, all of the above units with the suffixes 2 and 3 exist, and
    refer to the unit to the power of 2 and 3, e.g. ``metre3 = metre**3``.
    
    Arithmetic
    ~~~~~~~~~~
    
    The following operations on :class:`Quantity` objects require that the operands
    have the same dimensions:
    
        +, -, <, <=, >, >=, ==, !=
    
    The following operations on :class:`Quantity` objects work with any pair of
    operands:
    
        / and *
    
    In addition, ``-x`` and ``abs(x)`` will work on any :class:`Quantity` ``x``, and will return
    values with the same dimension as their argument.
    
    The power operation ``x**y`` requires that ``y`` be dimensionless.
    
    Casting
    ~~~~~~~

    The three rules that define the casting operations for
    :class:`Quantity` object are:
    
    1.  :class:`Quantity` op :class:`Quantity` = :class:`Quantity`:
        Performs dimension consistency check if appropriate.
    2.  Scalar op :class:`Quantity` = :class:`Quantity`: 
        Assumes that the scalar is dimensionless
    3.  other op :class:`Quantity` = other:
        The :class:`Quantity` object is downcast to a ``float``
    
    Scalar types are 1 dimensional number types, including ``float``, ``int``, etc.
    but not ``array``.
    
    The :class:`Quantity` class is a derived class of ``float``, so many other operations
    will also downcast to ``float``. For example, ``sin(x)`` where ``x`` is a quantity
    will return ``sin(float(x))`` without doing any dimension checking. Although
    see the Brian.unitsafefunctions module for a way round this. It is better
    to be explicit about casting if you can be.
    
    TODO: more details on ``numpy.array``/:class:`Quantity` operations?
    
    :func:`check_units` decorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    The decorator :func:`check_units` can be used to check that the arguments passed
    to a function have the right units, e.g.::
    
        @check_units(I=amp,V=volt)
        def find_resistance(I,V):
            return V/I
    
    will work if you try ``find_resistance(1*amp,1*volt)`` but raise an exception
    if you try ``find_resistance(1,1)`` say.
    """
    reinit_default_clock()

    # the following units should exist:
    units_which_should_exist = [ metre, meter, kilogram, second, amp, kelvin, mole, candle,
                                radian, steradian, hertz, newton, pascal, joule, watt,
                                coulomb, volt, farad, ohm, siemens, weber, tesla, henry,
                                celsius, lumen, lux, becquerel, gray, sievert, katal,
                                gram, gramme ]
    # scaled versions of all these units should exist (we just check farad as an example)
    some_scaled_units = [ Yfarad, Zfarad, Efarad, Pfarad, Tfarad, Gfarad, Mfarad, kfarad,
                         hfarad, dafarad, dfarad, cfarad, mfarad, ufarad, nfarad, pfarad,
                         ffarad, afarad, zfarad, yfarad ]
    # and check that the above is in descending order of size
    import copy
    sorted_scaled_units = copy.copy(some_scaled_units)
    sorted_scaled_units.sort(reverse=True)
    assert some_scaled_units == sorted_scaled_units
    # some powered units
    powered_units = [ cmetre2, Yfarad3 ]

    # check that operations requiring consistent units work with consistent units
    a = 1 * kilogram
    b = 2 * kilogram
    c = [ a + b, a - b, a < b, a <= b, a > b, a >= b, a == b, a != b ]
    # check that given inconsistent units they raise an exception
    from operator import add, sub, mul, div, lt, le, gt, ge, eq, ne
    tryops = [add, sub, lt, le, gt, ge, eq, ne]
    a = 1 * kilogram
    b = 1 * second
    def inconsistent_operation(a, b, op):
        return op(a, b)
    for op in tryops:
        assert_raises(DimensionMismatchError, inconsistent_operation, a, b, op)

    # check that operations not requiring consistent units work
    a = 1 * kilogram
    b = 1 * second
    c = [ a * b, a / b ]

    # check that - and abs give results with the same dimensions
    assert (-a).has_same_dimensions(a)
    assert abs(a).has_same_dimensions(a)
    assert - abs(a) < abs(a) # well why not, this should be true

    # check that pow requires the index to be dimensionless
    a = (1 * kilogram) ** 0.352
    def inconsistent_power(a, b):
        return a ** b
    a = 1 * kilogram
    b = 1 * second
    assert_raises(DimensionMismatchError, inconsistent_power, a, b)

    # check casting rule 1
    a = 1 * kilogram
    b = 2 * kilogram
    for op in [add, sub, mul, div]: assert isinstance(op(a, b), Quantity)

    # check casting rule 2
    a = 1
    b = 1 * kilogram
    for op in [mul, div]: assert isinstance(op(a, b), Quantity) and isinstance(op(b, a), Quantity)
    for op in [add, sub]:
        assert_raises(DimensionMismatchError, inconsistent_operation, a, b, op)
    for op in [add, sub, mul, div]: assert isinstance(op(a, b / b), Quantity) and isinstance(op(b / b, a), Quantity)

    # check casting rule 3
    assert isinstance(numpy.array([1, 2]) * (1 * kilogram), numpy.ndarray)
    assert isinstance((1 * kilogram) * numpy.array([1, 2]), numpy.ndarray)

    # check_units decorator
    @check_units(I=amp, V=volt)
    def find_resistance(I, V):
        return V / I
    R = find_resistance(1 * amp, 1 * volt)
    assert_raises(DimensionMismatchError, find_resistance, 1, 1)

if __name__ == '__main__':
    test()



# Neural Cortical Model of a single neuron with poisson input
from brian import *
from brian.library.synapses import *
import time
import random

defaultclock.dt=0.01*ms

vr    = -90*mvolt   # reset/resting membrane potential
vt    = -55*mvolt   # threshold
ref   =  1*msecond  # refractory period

tauge = 0.25*msecond # excitatory conductance time constant

El    = -70*mvolt   # resting potential
Ee    =   0*mvolt   # reversal potential of excitatory ionic channels

gl    = 10.*nsiemens # leak conductance

#compute the somatic and dendritic capacitance for each neuron type
Rmse = 600*Mohm
taue  =  10*msecond # somatic excitatory membrane time constant
Ces = taue/Rmse # excitatory cell somatic capacitance

#Poisson input value
Pinput = 150.5*nS # Highly exaggerated!
Pfreq  = 50*kHz

# Number of neurons
N   = 1

###### Neuron model
exc_soma = MembraneEquation(C=Ces) +  Current('I=gl*(El-vm):amp')
exc_soma += alpha_conductance(input='ge' ,E=Ee, tau=tauge)

print exc_soma

# Create Excitatory Neural Sheet
NG  = NeuronGroup(N, model=exc_soma, threshold='vm>vt',reset=vr,refractory=ref)

###### Make the Connections
#Poisson Input
PG  = PoissonGroup(N, N*[Pfreq])

P = Connection(PG, NG, 'ge')
P.connect_one_to_one(PG, NG, weight=Pinput)

SpP = SpikeMonitor(PG)
SpNG = SpikeMonitor(NG)

M=StateMonitor(NG,'vm',record=True)

#status
print 'Running...'
run(1*second)

###### analysis output

print SpP.nspikes, '\tpoisson spikes'
print SpNG.nspikes, '\ttotal spikes'

subplot(221)
raster_plot(SpP)
ylabel('Poisson Neuron Id')

subplot(223)
raster_plot(SpNG)
ylabel('Neuron Id')

subplot(222)
plot(M.times/ms,M[0]/mV)

show()



from codegen import *
from integration_schemes import *
from codegen_python import *
from codegen_c import *
from codegen_gpu import *

if __name__ == '__main__':
    if True:
        from brian import *
        eqs = Equations('''
        dV/dt = -W*V/(10*second) : volt 
        dW/dt = -V**2/(1*second) : volt
        ''')
        scheme = exp_euler_scheme
        print 'Equations'
        print '========='
        print eqs
        print 'Scheme'
        print '======'
        for block_specifier, block_code in scheme:
            print block_specifier
            print block_code
        print 'Python code'
        print '==========='
        print PythonCodeGenerator().generate(eqs, scheme)
        print 'C code'
        print '======'
        print CCodeGenerator().generate(eqs, scheme)
        print 'GPU code'
        print '======'
        print GPUCodeGenerator().generate(eqs, scheme)



"""
This example demonstrates the PoissonGroup object. Here we have
used a custom function to generate different rates at different
times.

This example also demonstrates a custom SpikeMonitor.
"""
#import brian_no_units # uncomment to run faster
from brian import *

# Rates

r1 = arange(101, 201) * 0.1 * Hz
r2 = arange(1, 101) * 0.1 * Hz

def myrates(t):
    if t < 10 * second:
        return r1
    else:
        return r2
# More compact: myrates=lambda t: (t<10*second and r1) or r2

# Neuron group
P = PoissonGroup(100, myrates)

# Calculation of rates

ns = zeros(len(P))

def ratemonitor(spikes):
    ns[spikes] += 1

Mf = SpikeMonitor(P, function=ratemonitor)
M = SpikeMonitor(P)

# Simulation and plotting

run(10 * second)
print "Rates after 10s:"
print ns / (10 * second)

ns[:] = 0
run(10 * second)
print "Rates after 20s:"
print ns / (10 * second)

raster_plot()
show()



'''
A complicated example of using multiprocessing for multiple runs of a simulation
with different parameters, using a GUI to monitor and control the runs.

This example features:

* An indefinite number of runs, with a set of parameters for each run generated
  at random for each run.
* A plot of the output of all the runs updated as soon as each run is completed.
* A GUI showing how long each process has been running for and how long until
  it completes, and with a button allowing you to terminate the runs.

A simpler example is in ``examples/multiprocessing/multiple_runs_simple.py``.
'''

# We use Tk as the backend for the GUI and matplotlib so as to avoid any
# threading conflicts
import matplotlib
matplotlib.use('TkAgg')

from brian import *
import Tkinter, time, multiprocessing, os
from brian.utils.progressreporting import make_text_report
from Queue import Empty as QueueEmpty


class SimulationController(Tkinter.Tk):
    '''
    GUI, uses Tkinter and features a progress bar for each process, and a callback
    function for when the terminate button is clicked.
    '''
    def __init__(self, processes, terminator, width=600):
        Tkinter.Tk.__init__(self, None)
        self.parent = None
        self.grid()
        button = Tkinter.Button(self, text='Terminate simulation',
                                command=terminator)
        button.grid(column=0, row=0)
        self.pb_width = width
        self.progressbars = []
        for i in xrange(processes):
            can = Tkinter.Canvas(self, width=width, height=30)
            can.grid(column=0, row=1 + i)
            can.create_rectangle(0, 0, width, 30, fill='#aaaaaa')
            r = can.create_rectangle(0, 0, 0, 30, fill='#ffaaaa', width=0)
            t = can.create_text(width / 2, 15, text='')
            self.progressbars.append((can, r, t))
        self.results_text = Tkinter.Label(self, text='Computed 0 results, time taken: 0s')
        self.results_text.grid(column=0, row=processes + 1)
        self.title('Simulation control')

    def update_results(self, elapsed, complete):
        '''
        Method to update the total number of results computed and the amount of time taken.
        '''
        self.results_text.config(text='Computed ' + str(complete) + ', time taken: ' + str(int(elapsed)) + 's')
        self.update()

    def update_process(self, i, elapsed, complete, msg):
        '''
        Method to update the status of a given process.
        '''
        can, r, t = self.progressbars[i]
        can.itemconfigure(t, text='Process ' + str(i) + ': ' + make_text_report(elapsed, complete) + ': ' + msg)
        can.coords(r, 0, 0, int(self.pb_width * complete), 30)
        self.update()

def sim_mainloop(pool, results, message_queue):
    '''
    Monitors results of a simulation as they arrive
    
    pool is the multiprocessing.Pool that the processes are running in,
    results is the AsyncResult object returned by Pool.imap_unordered which
    returns simulation results asynchronously as and when they are ready,
    and message_queue is a multiprocessing.Queue used to communicate between
    child processes and the server process. In this case, we use this Queue to
    send messages about the percent complete and time elapsed for each run.
    '''
    # We use this to enumerate the processes, mapping their process IDs to an int
    # in the range 0:num_processes.
    pid_to_id = dict((pid, i) for i, pid in enumerate([p.pid for p in pool._pool]))
    num_processes = len(pid_to_id)
    start = time.time()
    stoprunningsim = [False]
    # This function terminates all the pool's child processes, it is used as
    # the callback function called when the terminate button on the GUI is clicked.
    def terminate_sim():
        pool.terminate()
        stoprunningsim[0] = True
    controller = SimulationController(num_processes, terminate_sim)
    for i in range(num_processes):
        controller.update_process(i, 0, 0, 'no info yet')
    i = 0
    while True:
        try:
            # If there is a new result (the 0.1 means wait 0.1 seconds for a
            # result before giving up) then this try clause will execute, otherwise
            # a TimeoutError will occur and the except clause afterwards will
            # execute.
            weight, numspikes = results.next(0.1)
            # if we reach here, we have a result to plot, so we plot it and
            # update the GUI
            plot_result(weight, numspikes)
            i = i + 1
            controller.update_results(time.time() - start, i)
        except multiprocessing.TimeoutError:
            # if we're still waiting for a new result, we can process events in
            # the message_queue and update the GUI if there are any.
            while not message_queue.empty():
                try:
                    # messages here are of the form: (pid, elapsed, complete)
                    # where pid is the process ID of the child process, elapsed
                    # is the amount of time elapsed, and complete is the
                    # fraction of the run completed. See function how_many_spikes
                    # to see where these messages come from.
                    pid, elapsed, complete = message_queue.get_nowait()
                    controller.update_process(pid_to_id[pid], elapsed, complete, '')
                except QueueEmpty:
                    break
            controller.update()
            if stoprunningsim[0]:
                print 'Terminated simulation processes'
                break
    controller.destroy()

def plot_result(weight, numspikes):
    plot([weight], [numspikes], '.', color=(0, 0, 0.5))
    axis('tight')
    draw() # this forces matplotlib to redraw

# Note that how_many_spikes only takes one argument, which is a tuple of
# its actual arguments. The reason for this is that Pool.imap_unordered can only
# pass a single argument to the function its applied to, but that argument can
# be a tuple...
def how_many_spikes((excitatory_weight, message_queue)):
    reinit_default_clock()
    clear(True)

    eqs = '''
    dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
    dge/dt = -ge/(5*ms) : volt
    dgi/dt = -gi/(10*ms) : volt
    '''
    P = NeuronGroup(4000, eqs, threshold= -50 * mV, reset= -60 * mV)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(3200)
    Pi = P.subgroup(800)
    Ce = Connection(Pe, P, 'ge')
    Ci = Connection(Pi, P, 'gi')
    Ce.connect_random(Pe, P, 0.02, weight=excitatory_weight)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)
    M = SpikeMonitor(P)

    # This reporter function is called every second, and it sends a message to
    # the server process updating the status of the current run.
    def reporter(elapsed, complete):
        message_queue.put((os.getpid(), elapsed, complete))

    run(4000 * ms, report=reporter, report_period=1 * second)

    return (excitatory_weight, M.nspikes)


if __name__ == '__main__':
    numprocesses = None # number of processes to use, set to None to have one per CPU
    # We have to use a Queue from the Manager to send messages from client
    # processes to the server process
    manager = multiprocessing.Manager()
    message_queue = manager.Queue()
    pool = multiprocessing.Pool(processes=numprocesses)
    # This generator function repeatedly generates random sets of parameters
    # to pass to the how_many_spikes function
    def args():
        while True:
            weight = rand()*3.5 * mV
            yield (weight, message_queue)
    # imap_unordered returns an AsyncResult object which returns results as
    # and when they are ready, we pass this results object which is returned
    # immediately to the sim_mainloop function which monitors this, updates the
    # GUI and plots the results as they come in.
    results = pool.imap_unordered(how_many_spikes, args())
    ion() # this puts matplotlib into interactive mode to plot as we go
    sim_mainloop(pool, results, message_queue)



'''
Hodgkin-Huxley model
Assuming area 1*cm**2
'''

from brian import *
from brian.library.ionic_currents import *

#defaultclock.dt=.01*ms # more precise
El = 10.6 * mV
EK = -12 * mV
ENa = 120 * mV
eqs = MembraneEquation(1 * uF) + leak_current(.3 * msiemens, El)
eqs += K_current_HH(36 * msiemens, EK) + Na_current_HH(120 * msiemens, ENa)
eqs += Current('I:amp')

neuron = NeuronGroup(1, eqs, implicit=True, freeze=True)

trace = StateMonitor(neuron, 'vm', record=True)

run(100 * ms)
neuron.I = 10 * uA
run(100 * ms)
plot(trace.times / ms, trace[0] / mV)
show()



"""
Separation of an array of synapses into a list of arrays with no
repeated postsynaptic neurons
"""
from numpy import *
from time import time

'''
Testing whether we have repeated indexes.

OPTION 1
--------
x=all(bincount(post)<2)

This is very slow with bincount if the number of synapses is large.
And it's a memory expensive way to do it. In principle, we don't need
memory allocation for this test.

OPTION 2
--------
x=len(unique(post))<len(post)

This is much faster and less memory expensive. But it's still significant
(about 0.25 s for the CUBA example).
'''
if False:
    Nsyn=100000
    post=random.randint(Nsyn,size=160) # post neurons
    t1=time()
    for _ in range(10000): # 1s with dt=0.1 ms 
        #x=all(bincount(post)<2)
        x=len(unique(post))<len(post)
    t2=time()
    print t2-t1

'''
Dispatching.

We just use unique repeatedly. It returns indexes of the first occurrence of
each post neuron. We put a flag saying that we processed these indexes and
we iterate.
(Note that here we could save one iteration)

The code prints synapses and corresponding post neurons at each iteration. This
would be replaced by executing the pre or post code in Synapses.
'''
if False:
    synapses=array([1,2,3,4,5,6,7,8,9],dtype=int) # numbers don't matter
    post=array([7,1,3,4,1,3,6,1,2],dtype=int) # post neurons

    #k=0 # number of iterations
    u,i=unique(post,return_index=True)
    print synapses[i],post[i]
    while (len(u)<len(post)) & (len(u)>1):
        post[i]=-1 # flag meaning we've processed them
        u,i=unique(post,return_index=True)
        #k+=1
        print synapses[i[1:]],u[1:]

'''
Same thing as above, but in a way closer to the final code.

Seems to work (yay)
'''

spikes = [0, 1, 2, 3, 4]
pre = array([0, 1, 2, 2, 3, 3, 4, 4])
post = array([0, 1, 0, 1, 0, 0, 1, 1])

vout = zeros(2)
w = array([1, 2, 1, 2, 1, 2])

def update(weights, voltage, synapses):
    voltage[post[synapses]] += w[synapses]
    print voltage

for j in spikes:
    synapses = nonzero(pre == j)[0]
    post_neurons = post[synapses]
    
    u,i = unique(post_neurons, return_index=True)
    print 'base update'
    print 'concerned synapses', synapses
    print 'u, i', u, i
    update(w, vout, i)
    if len(u) < len(post_neurons): 
        while len(u) < len(post_neurons) and (post_neurons > -1).any():
            post_neurons[i] = -1
            u, i = unique(post_neurons, return_index = True)
            print 'repeated update'
            print 'concerned synapses', synapses
            print 'u, i', u, i
            update(w, vout, i[1:])



from ...optimiser import freeze
from string import Template
import re
from rewriting import *

__all__ = ['CodeGenerator']


class CodeGenerator(object):
    '''
    CodeGenerator base class for generating code in a variety of languages from mathematical expressions
    
    Base classes can define their own behaviour, but the following scheme is expected to be fairly
    standard.
    
    The initialiser takes at least one argument ``sympy_rewrite``, which defaults to ``True``, which
    specifies whether or not expressions should be simplified by passing them through sympy, so that
    e.g. ``V/(10*0.1)`` gets simplified to ``V*1.0`` (but not to ``V`` for some reason).
    
    There are also the following methods::
    
    .. method:: generate(eqs, scheme)
    
        Returns code in the target language for the set of equations ``eqs`` according to the
        integration scheme ``scheme``. The following schemes are included by default:
        
        * ``euler_scheme``
        * ``rk2_scheme``
        * ``exp_euler_scheme``
    
        This method may also call the :meth:`initialisation` and :meth:`finalisation` methods
        for code to include at the start and end. By default, it will call the :meth:`scheme`
        method to fill in the code in between.
        
    .. method:: initialisation(eqs)
    
        Returns initialisation code.
        
    .. method:: finalisation(eqs)
    
        Returns finalisation code.
    
    .. method:: scheme(eqs, scheme)
    
        Returns code for the given set of equations integrated according to the given scheme.
        See below for an explanation of integration schemes. In the process of generating
        code from the schemes, this method may call any of the following methods:
        :meth:`single_statement` to convert a single statement into code;
        :meth:`single_expr` to convert a single expression into code;
        :meth:`substitute` to substitute different values for specific variable names in
        given expressions; :meth:`vartype` to get the variable type specifier if necessary
        for that language.
    
    .. method:: single_statement(expr)
    
        Convert the given statement into code. By default, this looks for statements of the
        form ``A = B``, ``A += B``, etc. and applies the method :meth:`single_expr` to ``B``.
        Classes deriving from the base class typically only need to write a :meth:`single_expr`
        method and not a :meth:`single_statement` one.
    
    .. method:: single_expr(expr)
    
        Convert the given expression into code. For example, for Python you can do nothing, but
        for C, for example the expression ``A**B`` should be converted to ``pow(A, B)``, and
        this is done using sympy replacement. By default, this method will simplify the
        expression using sympy if the code generator was initialised with ``sympy_rewrite=True``.
        
    .. method:: substitute(var_expr, substitutions)
    
        Makes the given substitutions into the expression. ``substitutions`` should be a
        dictionary with keys the names of the variables to substitute, and values the
        value to substitute for that expression. By default these values are passed to
        the :meth:`single_substitute` method to check if there are any language specific
        things that need to be done to it, for example something substituting ``1`` should
        be replaced by ``ones(num_neurons)`` for Python which is vectorised. Typically
        this method shouldn't need to be rewritten for derived classes, whereas
        :meth:`single_substitute` might.
    
    .. method:: single_substitute(s)
    
        Replaces ``s``, a substitution value from :meth:`substitute`, with a language
        specific version if necessary. Should return a string.
    
    **Schemes**

    A scheme consists of a sequence of pairs ``(block_specifier, block_code)``. The
    ``block_specifier`` is currently unused, the schemes available at the moment all use
    ``('foreachvar', 'all')`` in this space - other specifiers are possible, e.g.
    ``'all_nonzero'`` instead of ``'all'``. ``block_code`` is a multi-line template string
    expressing that stage of the scheme (templating explained below). For each specifier,
    code pair, a block of code is generated in that order - multiple pairs can be used for
    separating stages in an integration scheme (i.e. do something for all variables, then
    do something else for all variables, etc.).
    
    Templating has two features, the first is standard Python templating replacements, e.g.
    the template string ``'${var}__buf'`` would be replaced by ``'x__buf'`` where the Python
    variable ``var='x'``. The available variables are:
    
    ``var``
        The current variable in the loop over each variable.
    ``var_expr``
        The right hand side of the differential equation, f(x) for the equation dx/dt=f(x).
    ``vartype``
        The data type (or ``''`` if none). Should be used at the start of the statement if
        the statement is declaring a new variable.
    
    The other feature of templating is that expressions of the form ``'@func(args)'`` will
    call the method ``func`` of the code generator with the given ``args`` and be replaced by
    that expression. Typically, this is used for the :meth:`substitute` method.
    
    Example scheme::

        rk2_scheme = [
            (('foreachvar', 'all'),
                """
                $vartype ${var}__buf = $var_expr
                $vartype ${var}__half = $var+dt*${var}__buf
                """),
            (('foreachvar', 'all'),
                """
                ${var}__buf = @substitute(var_expr, {var:var+'__buf'})
                $var += dt*${var}__buf
                """)
            ]
    '''
    def __init__(self, sympy_rewrite=True):
        self.sympy_rewrite = sympy_rewrite

    def single_statement(self, expr):
        m = re.search(r'[^><=]=', expr)
        if m:
            return expr[:m.end()] + ' ' + self.single_expr(expr[m.end():])
        return expr

    def single_expr(self, expr):
        if self.sympy_rewrite is not False:
            if self.sympy_rewrite is True:
                #rewriters = [floatify_numbers]
                rewriters = []
            else:
                rewriters = self.sympy_rewrite
            return sympy_rewrite(expr.strip(), rewriters)
        return expr.strip()

    def vartype(self):
        return ''

    def initialisation(self, eqs):
        return ''

    def finalisation(self, eqs):
        return ''

    def generate(self, eqs, scheme):
        code = self.initialisation(eqs)
        code += self.scheme(eqs, scheme)
        code += self.finalisation(eqs)
        return code

    def scheme(self, eqs, scheme):
        code = ''
        all_variables = eqs._eq_names + eqs._diffeq_names + eqs._alias.keys() + ['t']
        vartype = self.vartype()
        for block_specifier, block_code in scheme:
            # for the moment, processing of block_specifier is very crude
            if block_specifier == ('foreachvar', 'all'):
                vars_to_use = eqs._diffeq_names
            elif block_specifier == ('foreachvar', 'nonzero'):
                vars_to_use = eqs._diffeq_names_nonzero
            for line in block_code.split('\n'):
                line = line.strip()
                if line:
                    origline = line
                    for var in vars_to_use:
                        vars = eqs._diffeq_names
                        line = origline
                        namespace = eqs._namespace[var]
                        var_expr = freeze(eqs._string[var], all_variables, namespace)
                        while 1:
                            m = re.search(r'\@(\w+)\(', line)
                            if not m:
                                break
                            methname = m.group(1)
                            start, end = m.span()
                            numopen = 1
                            for i in xrange(end, len(line)):
                                if line[i] == '(':
                                    numopen += 1
                                if line[i] == ')':
                                    numopen -= 1
                                if numopen == 0:
                                    break
                            if numopen != 0:
                                raise SyntaxError('Parentheses unmatching.')
                            args = line[start + 1:i + 1]
                            #print args
                            exec 'line = line[:start]+self.' + args + '+line[i+1:]'
                        substitutions = {'vartype':vartype,
                                         'var':var,
                                         'var_expr':var_expr}
                        t = Template(line)
                        code += self.single_statement(t.substitute(**substitutions)) + '\n'
        return code

    def single_substitute(self, s):
        return str(s)

    def substitute(self, var_expr, substitutions):
        for var, replace_var in substitutions.iteritems():
            var_expr = re.sub(r'\b' + var + r'\b', self.single_substitute(replace_var), var_expr)
        return var_expr



from brian import *
import time
import pickle
from operator import itemgetter
from cuba_runopts import *

def cubanetwork(N, Nsyn=80., we=1.62 * mV):
    eqs = '''
    dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
    dge/dt = -ge/(5*ms) : volt
    dgi/dt = -gi/(10*ms) : volt
    '''

    Ne = int(N * 0.8)
    Ni = N - Ne

    P = NeuronGroup(N, model=eqs,
                  threshold= -50 * mV, reset= -60 * mV, refractory=5 * ms)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(Ne)
    Pi = P.subgroup(Ni)

    if cuba_opts['connections']:
        Ce = Connection(Pe, P, 'ge')
        Ci = Connection(Pi, P, 'gi')
        cp = Nsyn / N
        Ce.connect_random(Pe, P, cp, weight=we)
        Ci.connect_random(Pi, P, cp, weight= -9 * mV)
#        from brian.experimental.stdp_sparse import SparseSTDPConnectionMatrix
#        Ce.W = SparseSTDPConnectionMatrix(Ce.W)
#        Ci.W = SparseSTDPConnectionMatrix(Ci.W)
    else:
        Ce = None
        Ci = None

    M = PopulationSpikeCounter(P)

    if cuba_opts['connections']:
        net = Network(P, Ce, Ci, M)
    else:
        net = Network(P, M)
    net.run(0 * ms)
    net.Nsyn = Nsyn
    net.we = we

    return P, Pe, Pi, Ce, Ci, M, net

def cuba(P, Pe, Pi, Ce, Ci, M, net):
    reinit_default_clock()
    P.v = -60 * mV + 10 * mV * rand(len(P))
    if cuba_opts['connections']:
        Ce.__init__(Pe, P, 'ge')
        Ci.__init__(Pi, P, 'gi')
        cp = net.Nsyn / len(P)
        Ce.connect_random(Pe, P, cp, weight=net.we)
        Ci.connect_random(Pi, P, cp, weight= -9 * mV)
        Ce.W.freeze()
        Ci.W.freeze()
    M.reinit()
    tstart = time.time()
    net.run(duration)
    tend = time.time()
    print '.',
    return tend - tstart, M.nspikes

def cuba_average(N, repeats, best, Nsyn=80., we=1.62 * mV):
    cubadata = cubanetwork(N, Nsyn, we)
    cubaruns = [cuba(*cubadata) for i in range(repeats)]
    print
    print 'Finished run', N, '-', Nsyn, '-', we
    cubaruns.sort(key=itemgetter(0))
    cubaruns = cubaruns[:best]
    t, ns = zip(*cubaruns)
    t = array(t)
    ns = array(ns)
    return N, mean(t), mean(ns), cubaruns

def cuba_runs(Nvals, repeats, best):
    return [cuba_average(N, repeats, best) for N in Nvals]

def cuba_runs_varyconnectivity(N, Nsynvals, repeats, best):
    return [cuba_average(N, repeats, best, Nsyn) for Nsyn in Nsynvals]

def cuba_runs_varywe(N, wevals, repeats, best):
    return [cuba_average(N, repeats, best, we=we) for we in wevals]

def do_runs(fname='data/brian_cuba_results.pkl'):
    ca = cuba_runs(Nvals, repeats, best)
    output = open(fname, 'wb')
    pickle.dump(ca, output, -1)
    output.close()
    return ca

def do_runs_varyconnectivity(fname='data/brian_cuba_varycon_results.pkl'):
    ca = cuba_runs_varyconnectivity(N_varycon, Nsynvals, repeats, best)
    output = open(fname, 'wb')
    pickle.dump(ca, output, -1)
    output.close()
    return ca

def do_runs_varywe(fname='data/brian_cuba_varywe_results.pkl'):
    ca = cuba_runs_varywe(N_varywe, wevals, repeats, best)
    output = open(fname, 'wb')
    pickle.dump(ca, output, -1)
    output.close()
    return ca

if __name__ == '__main__':
    #print do_runs()
    #print do_runs_varyconnectivity()
    X = cubanetwork(16000, we=9 * mV)
    print 'Built X'
    print cuba(*X)



"""
Event-based STDP

If this is satisfactory, we could merge it into
the usual STDP class. New features:
* t is in the namespace
* there are synaptic variables (matrices instead of 1D arrays)

If we want heterogeneous delays, we will need 3D arrays for synaptic
variables!

We may also want to insert the fast exp in the namespace.

A few propositions:
* Give easy access to pre/post synaptic variables in neurongroups,
for example: v_post
* Automatic handling of t_pre/t_post (time of last update of variables)
* Specification of pre/post variables in equations, or automatic detection, e.g.:
A_pre : 1 (pre)
A_post : 1 (post)
A : 1 (synapse) # or other keyword?
"""
from brian import *
from brian.utils.documentation import flattened_docstring
from brian.inspection import *
from brian.equations import *
from brian.optimiser import freeze
import re

__all__=['EventBasedSTDP']

class EventSTDPUpdater(SpikeMonitor):
    '''
    Updates STDP variables at spike times
    Almost the same as usual one, but with t in the namespace
    '''
    def __init__(self, source, C, code, namespace, delay=0 * ms):
        '''
        source = source group
        C = connection
        vars = variable names
        M = matrix of the linear differential system
        code = code to execute for every spike
        namespace = namespace for the code
        delay = transmission delay 
        '''
        super(EventSTDPUpdater, self).__init__(source, record=False, delay=delay)
        self._code = code # update code
        self._namespace = namespace # code namespace
        self.C = C

    def propagate(self, spikes):
        if len(spikes):
            self._namespace['spikes'] = spikes
            self._namespace['w'] = self.C.W
            self._namespace['t'] = self.C.source.clock._t
            exec self._code in self._namespace

class EventBasedSTDP(NetworkOperation):
    '''
    Spike-timing-dependent plasticity
    Event-based implementation
    '''
    def __init__(self, C, eqs, pre, post, wmin=0, wmax=Inf, level=0, clock=None, delay_pre=None, delay_post=None):
        '''
        C: connection object
        eqs: equations (with units)
        pre: Python code for presynaptic spikes
        post: Python code for postsynaptic spikes
        wmax: maximum weight (default unlimited)
        delay_pre: presynaptic delay
        delay_post: postsynaptic delay (backward propagating spike)
        '''
        NetworkOperation.__init__(self, lambda:None, clock=clock)
        # Convert to equations object
        if isinstance(eqs, Equations):
            eqs_obj = eqs
        else:
            eqs_obj = Equations(eqs, level=level + 1)
        # handle multi-line pre, post equations and multi-statement equations separated by ;
        if '\n' in pre:
            pre = flattened_docstring(pre)
        elif ';' in pre:
            pre = '\n'.join([line.strip() for line in pre.split(';')])
        if '\n' in post:
            post = flattened_docstring(post)
        elif ';' in post:
            post = '\n'.join([line.strip() for line in post.split(';')])

        # Check units
        eqs_obj.compile_functions()
        eqs_obj.check_units()
        # Get variable names
        vars = eqs_obj._diffeq_names

        # Create namespaces for pre and post codes
        pre_namespace = namespace(pre, level=level + 1)
        post_namespace = namespace(post, level=level + 1)
        pre_namespace['clip'] = clip
        post_namespace['clip'] = clip

        # freeze pre and post (otherwise units will cause problems) [why?]
        all_vars = list(vars) + ['w','t','t_pre','t_post']
        pre = '\n'.join(freeze(line.strip(), all_vars, pre_namespace) for line in pre.split('\n'))
        post = '\n'.join(freeze(line.strip(), all_vars, post_namespace) for line in post.split('\n'))

        # Create synaptic variables
        self.var=dict()
        #C.compress()
        for x in vars:
            self.var[x]=C.W.connection_matrix(copy=True)
            self.var[x][:]=0 # reset values
        
        # Create code
        # Indent and loop
        pre = re.compile('^', re.M).sub('    ', pre)
        post = re.compile('^', re.M).sub('    ', post)
        pre = 'for _i in spikes:\n' + pre
        post = 'for _i in spikes:\n' + post

        # Pre/post code
        for var in vars+['w']: # presynaptic variables (vectorisation)
            pre = re.sub(r'\b' + var + r'\b', var + '[_i,:]', pre)
            post = re.sub(r'\b' + var + r'\b', var + '[:,_i]', post)

        # Bounds: add one line to pre/post code (clip(w,min,max,w))
        # or actual code? (rather than compiled string)
        pre += '\n    w[_i,:]=clip(w[_i,:],%(min)e,%(max)e)' % {'min':wmin, 'max':wmax}
        post += '\n    w[:,_i]=clip(w[:,_i],%(min)e,%(max)e)' % {'min':wmin, 'max':wmax}
                
        log_debug('brian.stdp', 'PRE CODE:\n'+pre)
        log_debug('brian.stdp', 'POST CODE:\n'+post)
        # Compile code
        pre_code = compile(pre, "Presynaptic code", "exec")
        post_code = compile(post, "Postsynaptic code", "exec")

        # Delays
        connection_delay = C.delay * C.source.clock.dt
        if (delay_pre is None) and (delay_post is None): # same delays as the Connnection C
            delay_pre = connection_delay
            delay_post = 0 * ms
        elif delay_pre is None:
            delay_pre = connection_delay - delay_post
            if delay_pre < 0 * ms: raise AttributeError, "Postsynaptic delay is too large"
        elif delay_post is None:
            delay_post = connection_delay - delay_pre
            if delay_post < 0 * ms: raise AttributeError, "Postsynaptic delay is too large"

        # Put variables in namespace
        for ns in (pre_namespace,post_namespace):
            for var in vars:
                ns[var]=self.var[var]
        
        # create forward and backward Connection objects or SpikeMonitor objects
        pre_updater = EventSTDPUpdater(C.source, C, code=pre_code, namespace=pre_namespace, delay=delay_pre)
        post_updater = EventSTDPUpdater(C.target, C, code=post_code, namespace=post_namespace, delay=delay_post)
        self.contained_objects += [pre_updater, post_updater]

if __name__ == '__main__':
    pass



#
# Triplet rule STDP
#
# UNDER DEVELOPMENT - TESTING THE TRIPLET RULE IMPLEMENTATION
#
#
from brian import *

rate = 0.050 *kHz # Spike frequency
delta= 1 / rate # Interspike interval
N    = 60    # Number of spikes
t0   = 0 *ms # Time of the first spike
t1   = t0 + N * delta # Time of the last spike
tau  = 1*ms 

tau_plus  = 16.8 *ms
tau_x     = 101 *ms
tau_minus = 33.7 *ms
tau_y     = 125 *ms
A2_minus  = 7e-3
A3_minus  = 2.3e-4
A2_plus   = 5e-10
A3_plus   = 6.2e-3


presyn_train  = linspace(t0, t1, N, endpoint= False)
postsyn_train = presyn_train + 10*ms

input1 = SpikeGeneratorGroup(1, [(0, t * second) for t in presyn_train])
input2 = SpikeGeneratorGroup(1, [(0, t * second) for t in postsyn_train])

eqs = """
dv/dt = 0 * mV/ms  : mV
dw/dt = 0 /ms:1
"""

def myresetfunc(P, spikes):
 P.v[spikes] = H

SCR = SimpleCustomRefractoriness(myresetfunc, 0.01 *ms, state='v')

P = NeuronGroup(2, model=eqs, threshold="v > 5*mV", reset=0*mV)
P1 = P.subgroup(1)
P2 = P.subgroup(1)

C1 = Connection(input1, P1, 'v')
C1[0,0] = 10*mV
C2 = Connection(input2, P2, 'v')
C2[0,0] = 10*mV

C = Connection(P1, P2, 'w', weight=0)

eqs_stdp=Equations('''
dr1/dt = -r1/tau_plus  : 1
dr2/dt = -r2/tau_x     : 1
#www    = r1 - r2       : 1
do1/dt = -o1/tau_minus : 1
do2/dt = -o2/tau_y     : 1
#zzz    = o1 - o2       : 1
''')

#Triplet rule: fails to run
stdp=STDP(C,eqs=eqs_stdp,pre='r1+=1;w-=o1*(A2_minus+A3_minus*r2);r2+=1', post='o1+=1;w+=r1*(A2_plus+A3_plus*o2);o2+=1',wmax=100)#stdp=STDP(C,eqs=eqs_stdp,pre='r1+=1;w-=o1*(A2_minus+A3_minus*r2);r2+=1', post='o1+=1;w+=(A2_plus+A3_plus*o2);o2+=1',wmax=100)

#Modification: no mixing of pre and post terms on the weight update 
#stdp=STDP(C,eqs=eqs_stdp,pre='r1+=1;w-=o1*(A2_minus+A3_minus);r2+=1', post='o1+=1;w+=r1*(A2_plus+A3_plus);o2+=1',wmax=100)#stdp=STDP(C,eqs=eqs_stdp,pre='r1+=1;w-=o1*(A2_minus+A3_minus*r2);r2+=1', post='o1+=1;w+=(A2_plus+A3_plus*o2);o2+=1',wmax=100)


P.v = 0 *mV

M = StateMonitor(P, 'v', record=True)
run(600 *ms)
plot(M.times / ms, M[0] / mV, M.times / ms, M[1] / mV)
show()



from brian import *
from brian.experimental.modelfitting import *
from numpy.random import *
import time

if __name__ == '__main__':
    
    equations = Equations('''
        dV/dt=(R*I-V)/tau : 1
        I : 1
        R : 1
        tau : second
    ''')
    
    input = loadtxt('current.txt')
    trace = loadtxt('trace_artificial.txt')
    
    groups = 1
    overlap = 0*ms
#    input, trace = slice_trace(input, trace, slices = groups, overlap = overlap)
    
    
    neurons = 25600
    
    R = 3e9*ones(neurons)
    tau = 25*ms*ones(neurons)
    
    criterion = LpError(p=2, varname='V')
    
    t0 = time.clock()
    criterion_values = simulate( model = equations,
                                reset = 0,
                                threshold = 1,
                                data = trace,
                                input = input,
                                use_gpu = True,
                                groups = groups,
                                overlap = overlap,
                                stepsize = 128*ms,
                                dt = .1*ms,
                                criterion = criterion,
                                neurons = neurons,
                                R = R,
                                tau = tau)
    dur = time.clock()-t0
    print dur
    print min(criterion_values), max(criterion_values) 
    


from brian import *
from scipy.sparse import csr_matrix
import time

N = 1000
p = 0.1
repeats = 1000

G = NeuronGroup(N, 'V:1')
H = NeuronGroup(N, 'V:1')
Cs = Connection(G, H, 'V', sparseness=p, weight=1.0, structure='sparse')
Cd = Connection(G, H, 'V', sparseness=p, weight=1.0, structure='dense')
Cs.compress()
Ws = Cs.W
Wd = Cd.W

VG = asarray(G.V)
VH = asarray(H.V)

rowj, rowdata = Ws.rowj, Ws.rowdata

def f():
    VG[:] = 0
    VG[:] = 0
    for i in arange(0, N): # i is the source
        VG[rowj[i]] += VH[i]*rowdata[i]
start = time.time()
for _ in xrange(repeats):
    f()
print 'Time using loop:', time.time()-start

def f():
    VG[:] = 0
    VG[:] = 0
    for i in xrange(N): # i is the source
        VG[rowj[i]] += VH[i]*rowdata[i]
start = time.time()
for _ in xrange(repeats):
    f()
print 'Time using optimised loop:', time.time()-start

def f():
    VG[:] = dot(VH, Wd)
start = time.time()
for _ in xrange(repeats):
    f()
print 'Time using dense:', time.time()-start

Wss = csr_matrix((Ws.alldata, Ws.allj, Ws.rowind), (N, N))
def f():
    VG[:] = Wss*VH
start = time.time()
for _ in xrange(repeats):
    f()
print 'Time using scipy sparse:', time.time()-start



# -*- coding:utf-8 -*-
"""
Multiple synapses for each axon
"""
from brian import *

input=SpikeGeneratorGroup(2,[(0,10*ms),(1,50*ms)])
neurons=NeuronGroup(3,model='dv/dt=-v/(10*ms):1')
synapses=Connection(input,neurons,'v',delay=True,max_delay=10*ms)

M=StateMonitor(neurons,'v',record=True)

#synapses[0,0]=1
#synapses.delay[0,0]=2*ms

synapses.W.rows[0]=[1,1]
synapses.W.data[0]=[1,.5]
synapses.delayvec.rows[0]=[1,1]
synapses.delayvec.data[0]=[2*ms,5*ms]

run(80*ms)

plot(M.times/ms,M[0],'r')
plot(M.times/ms,M[1],'b')
plot(M.times/ms,M[2],'k')
show()



'''
Adapted from
Theory of Arachnid Prey Localization
W. Sturzl, R. Kempter, and J. L. van Hemmen
PRL 2000

Poisson inputs are replaced by integrate-and-fire neurons

Romain Brette
'''
from brian import *

# Parameters
degree = 2 * pi / 360.
duration = 500 * ms
R = 2.5 * cm # radius of scorpion
vr = 50 * meter / second # Rayleigh wave speed
phi = 144 * degree # angle of prey
A = 250 * Hz
deltaI = .7 * ms # inhibitory delay
gamma = (22.5 + 45 * arange(8)) * degree # leg angle
delay = R / vr * (1 - cos(phi - gamma))  # wave delay

# Wave (vector w)
t = arange(int(duration / defaultclock.dt) + 1) * defaultclock.dt
Dtot = 0.
w = 0.
for f in range(150, 451):
    D = exp(-(f - 300) ** 2 / (2 * (50 ** 2)))
    xi = 2 * pi * rand()
    w += 100 * D * cos(2 * pi * f * t + xi)
    Dtot += D
w = .01 * w / Dtot

# Rates from the wave
def rates(t):
    return w[array(t / defaultclock.dt, dtype=int)]

# Leg mechanical receptors
tau_legs = 1 * ms
sigma = .01
eqs_legs = """
dv/dt=(1+rates(t-d)-v)/tau_legs+sigma*(2./tau_legs)**.5*xi:1
d : second
"""
legs = NeuronGroup(8, model=eqs_legs, threshold=1, reset=0, refractory=1 * ms)
legs.d = delay
spikes_legs = SpikeCounter(legs)

# Command neurons
tau = 1 * ms
taus = 1 * ms
wex = 7
winh = -2
eqs_neuron = '''
dv/dt=(x-v)/tau : 1
dx/dt=(y-x)/taus : 1 # alpha currents
dy/dt=-y/taus : 1
'''
neurons = NeuronGroup(8, model=eqs_neuron, threshold=1, reset=0)
synapses_ex = IdentityConnection(legs, neurons, 'y', weight=wex)
synapses_inh = Connection(legs, neurons, 'y', delay=deltaI)
for i in range(8):
    synapses_inh[i, (4 + i - 1) % 8] = winh
    synapses_inh[i, (4 + i) % 8] = winh
    synapses_inh[i, (4 + i + 1) % 8] = winh
spikes = SpikeCounter(neurons)

run(duration)
nspikes = spikes.count
x = sum(nspikes * exp(gamma * 1j))
print "Angle (deg):", arctan(imag(x) / real(x)) / degree
polar(concatenate((gamma, [gamma[0] + 2 * pi])), concatenate((nspikes, [nspikes[0]])) / duration)
show()
